@@ -59,7 +59,12 @@
   (print-unreadable-object (type-class stream :type t)
     (prin1 (alien-type-class-name type-class) stream)))
 
-(defglobal *alien-type-classes* (make-hash-table :test 'eq)) ; keys are symbols
+;;; An benefit of EQL tables of symbols is that they will never rehash due to key
+;;; movement, because they use the symbol's hash which is constant even across
+;;; core restart, whereas EQ tables might rehash.
+;;; Perhaps we should consider using SXHASH on symbols for EQ tables.
+;;; And frankly, why isn't this particular table not a globaldb info instead?
+(defglobal *alien-type-classes* (make-hash-table)) ; keys are symbols
 
 (defun alien-type-class-or-lose (name)
   (or (gethash name *alien-type-classes*)
@@ -167,6 +172,17 @@
 (eval-when (:compile-toplevel :load-toplevel :execute)
   (create-alien-type-class-if-necessary 'root 'alien-type nil))
 
+(def!struct (alien-type
+             (:copier nil)
+             (:constructor make-alien-type
+                           (&key class bits alignment
+                            &aux (alignment
+                                  (or alignment (guess-alignment bits))))))
+  (class 'root :type symbol :read-only t)
+  (bits nil :type (or null unsigned-byte))
+  (alignment nil :type (or null unsigned-byte)))
+(!set-load-form-method alien-type (:xc :target))
+
 (defmethod print-object ((type alien-type) stream)
   (print-unreadable-object (type stream :type t)
     (sb-impl:print-type-specifier stream (unparse-alien-type type))))
@@ -514,7 +530,7 @@
   ;; the register uninitialized.  On those platforms, we use an
   ;; alien-rep of the full register width when checking for purposes
   ;; of return values and override the naturalize method to perform
-  ;; the sign extension (in compiler/target/c-call.lisp).
+  ;; the sign extension (in compiler/{arch}/c-call.lisp).
   (ecase context
     ((:normal #-(or x86 x86-64) :result)
      (list (if (alien-integer-type-signed type) 'signed-byte 'unsigned-byte)
@@ -1130,7 +1146,7 @@
   (arg-types (missing-arg) :type list)
   ;; The 3rd-party CFFI library uses presence of &REST in an argument list
   ;; as indicative of "..." in the C prototype. We can record that too.
-  (varargs nil :type (or boolean (eql :unspecified)))
+  (varargs nil :type (or boolean fixnum (eql :unspecified)))
   (stub nil :type (or null function))
   (convention nil :type calling-convention))
 ;;; The safe default is to assume that everything is varargs.
@@ -1148,6 +1164,7 @@
 ;;; result-type in this case. If convention is ever to become a part
 ;;; of result-type, such a syntax can be retained.
 
+
 (define-alien-type-translator function (result-type &rest arg-types
                                                     &environment env)
   (binding* (((bare-result-type calling-convention)
@@ -1155,14 +1172,21 @@
                 ((cons calling-convention *)
                  (values (second result-type) (first result-type)))
                 (t result-type)))
-             (varargs (eq (car (last arg-types)) '&rest)))
+             (varargs (or (eq (car (last arg-types)) '&rest)
+                          (position '&optional arg-types)))
+             (arg-types (if (integerp varargs)
+                            (remove '&optional arg-types)
+                            arg-types)))
     (make-alien-fun-type
      :convention calling-convention
      :result-type (let ((*values-type-okay* t))
                     (parse-alien-type bare-result-type env))
      :varargs (or varargs *alien-fun-type-varargs-default*)
-     :arg-types (mapcar (lambda (arg-type) (parse-alien-type arg-type env))
-                        (if varargs (butlast arg-types) arg-types)))))
+     :arg-types (mapcar (lambda (arg-type)
+                          (parse-alien-type arg-type env))
+                        (if (eql varargs t)
+                            (butlast arg-types)
+                            arg-types)))))
 
 (define-alien-type-method (fun :unparse) (type)
   `(function ,(let ((result-type
