@@ -17,7 +17,7 @@
 
 ;;; The DEFSTRUCT-DESCRIPTION structure holds compile-time information
 ;;; about a structure type.
-;;; It is defined prior to LAYOUT because a LAYOUT-INFO slot
+;;; It is defined prior to WRAPPER because WRAPPER-INFO
 ;;; is declared to hold a DEFSTRUCT-DESCRIPTION.
 (def!struct (defstruct-description
              (:conc-name dd-)
@@ -94,11 +94,25 @@
 
 ;;; Careful here: if you add more bits, then adjust the bit packing for
 ;;; 64-bit layouts which also store LENGTH + DEPTHOID in the same word.
-(defconstant +structure-layout-flag+         #b00000001)
-(defconstant +pathname-layout-flag+          #b00000010)
-(defconstant +condition-layout-flag+         #b00001000)
-(defconstant +pcl-object-layout-flag+        #b00010000)
-(defconstant sb-vm:lockfree-list-node-flag   #b01000000) ; exported for use in gc-private.h
+(defconstant +structure-layout-flag+         #b000000001)
+(defconstant +pathname-layout-flag+          #b000000010)
+(defconstant +pcl-object-layout-flag+        #b000000100)
+(defconstant +condition-layout-flag+         #b000001000)
+(defconstant +simple-stream-layout-flag+     #b000010000)
+(defconstant +file-stream-layout-flag+       #b000100000)
+(defconstant +string-stream-layout-flag+     #b001000000)
+(defconstant +stream-layout-flag+            #b010000000)
+(defconstant +sequence-layout-flag+          #b100000000)
+(defconstant layout-flags-mask #xffff) ; "strictly flags" bits from the packed field
+
+;;; the type of LAYOUT-DEPTHOID and LAYOUT-LENGTH values.
+;;; Each occupies two bytes of the %BITS slot when possible,
+;;; otherwise a slot unto itself.
+(def!type layout-depthoid () '(integer -1 #x7FFF))
+(def!type layout-length () '(integer 0 #xFFFF))
+(def!type layout-bitmap () 'integer)
+;;; ID must be an fixnum for either value of n-word-bits.
+(def!type layout-id () '(signed-byte 30))
 
 ;;; The LAYOUT structure is pointed to by the first cell of instance
 ;;; (or structure) objects. It represents what we need to know for
@@ -135,11 +149,11 @@
 
 ;;; 32-bit is not done yet. Three slots are still used, instead of two.
 
-;;; Maximum value of N in ANCESTOR_N. Couldn't come up with a better name.
-(defconstant sb-c::layout-inherits-max-optimized-depth 5)
-(sb-xc:defstruct (layout (:copier nil)
+#-metaspace
+(progn
+(sb-xc:defstruct (wrapper (:copier nil)
                          ;; Parsing DEFSTRUCT uses a temporary layout
-                         (:constructor make-temporary-layout
+                         (:constructor make-temporary-wrapper
                              (clos-hash classoid inherits &aux (invalid nil))))
 
   ;; A packed field containing the DEPTHOID, LENGTH, and FLAGS
@@ -190,172 +204,81 @@
   #-64-bit (length 0 :type layout-length) ; smaller than SB-INT:INDEX
   ;; If this layout has some kind of compiler meta-info, then this is
   ;; it. If a structure, then we store the DEFSTRUCT-DESCRIPTION here.
-  (info nil :type (or null defstruct-description))
+  ;; If this layout is for an object of metatype STANDARD-CLASS,
+  ;; then these are the EFFECTIVE-SLOT-DEFINITION metaobjects.
+  ;; The two are mutually exclusive.
+  (%info nil :type (or list defstruct-description))
   ;; EQUALP comparator for two instances with this layout
   ;; Could be the generalized function, or a type-specific one
   ;; if the defstruct was compiled in a policy of SPEED 3.
   (equalp-impl #'equalp-err :type (sfunction (t t) boolean) :read-only t)
-  ;; If this layout is for an object of metatype STANDARD-CLASS,
-  ;; these are the EFFECTIVE-SLOT-DEFINITION metaobjects.
-  (slot-list nil :type list)
   ;; Information about slots in the class to PCL: this provides fast
   ;; access to slot-definitions and locations by name, etc.
   ;; See MAKE-SLOT-TABLE in pcl/slots-boot.lisp for further details.
   (slot-table #(1 nil) :type simple-vector)
-  ;; inherited layouts or 0, only pertinent to structure classoids.
-  ;; There is no need to store the layout at depths 0 or 1
-  ;; since they're predetermined to be T and STRUCTURE-OBJECT.
-  (ancestor_2 0)
-  (ancestor_3 0)
-  (ancestor_4 0)
-  (ancestor_5 0))
-(declaim (freeze-type layout))
+  (id-word0 0 :type word)
+  (id-word1 0 :type word)
+  (id-word2 0 :type word)
+  #-64-bit (id-word3 0 :type word)
+  #-64-bit (id-word4 0 :type word)
+  #-64-bit (id-word5 0 :type word))
+(declaim (freeze-type wrapper)))
 
-(defun equalp-err (a b)
-  (bug "EQUALP ~S ~S" a b))
-
-;;; Applicable only if bit-packed (for 64-bit architectures)
-(defmacro pack-layout-flags (depthoid length flags)
-  `(logior (ash ,(or depthoid -1) (+ 32 sb-vm:n-fixnum-tag-bits))
-           (ash ,(or length 0) 16)
-           ,(or flags 0)))
-
-(defmacro type-dd-length (type-name)
-  (dd-length (find-defstruct-description type-name)))
-
-(defmacro get-dsd-index (type-name slot-name)
-  ;; It seems to be an error in CCL to declare something NOTINLINE
-  ;; if it is an unknown function.
-  #+sb-xc (declare (notinline dsd-index)) ; forward reference
-  (dsd-index (find slot-name
-                   (dd-slots (find-defstruct-description type-name))
-                   :key #'dsd-name)))
-
-(defmacro set-bitmap-from-layout (to-layout from-layout)
-  #+sb-xc-host (declare (ignore to-layout from-layout))
-  #-sb-xc-host
-  `(let ((index (type-dd-length layout)))
-     (dotimes (i (layout-bitmap-words ,from-layout))
-       (setf (%raw-instance-ref/word ,to-layout index)
-             (%raw-instance-ref/word ,from-layout index))
-       (incf index))))
-
-;;; See the pictures above DD-BITMAP in src/code/defstruct for the details.
-(defconstant standard-gf-primitive-obj-layout-bitmap
-  #+compact-instance-header  6
-  #-compact-instance-header -4)
-
-#+sb-xc-host
-(defmacro set-layout-inherits (layout inherits)
-  `(setf (layout-inherits ,layout) ,inherits))
-#-sb-xc-host
-(defmacro set-layout-inherits (layout inherits &optional recompute-bitmap)
-  `(let* ((l ,layout) (i ,inherits) (d (length i)))
-     (setf (layout-inherits l) i)
-     (setf (layout-ancestor_2 l) (if (> d 2) (svref i 2) 0)
-           (layout-ancestor_3 l) (if (> d 3) (svref i 3) 0)
-           (layout-ancestor_4 l) (if (> d 4) (svref i 4) 0)
-           (layout-ancestor_5 l) (if (> d 5) (svref i 5) 0))
-     ;; This part is for PCL where a class can forward-reference its superclasses
-     ;; and we only decide at class finalization time whether it is funcallable.
-     ;; Picking the right bitmap could probably be done sooner given the metaclass,
-     ;; but this approach avoids changing how PCL uses MAKE-LAYOUT.
-     ;; The big comment above MAKE-IMMOBILE-FUNINSTANCE in src/code/x86-64-vm
-     ;; explains why we differentiate between SGF and everything else.
-     ,(when recompute-bitmap
-        `(when (find ,(find-layout 'function) i)
-           (let ((bitmap
-                  #+immobile-code ; there are two possible bitmaps
-                  ;; *SGF-WRAPPER* isn't defined as yet, but this is just an s-expression.
-                  (if (or (find sb-pcl::*sgf-wrapper* i) (eq l sb-pcl::*sgf-wrapper*))
-                      standard-gf-primitive-obj-layout-bitmap
-                      +layout-all-tagged+)
-                  ;; there is only one possible bitmap otherwise
-                  #-immobile-code standard-gf-primitive-obj-layout-bitmap))
-             (setf (%raw-instance-ref/word l (type-dd-length layout))
-                   (logand sb-ext:most-positive-word bitmap)))))
-     l))
-(push '("SB-KERNEL" set-layout-inherits) *!removable-symbols*)
-
-;;; For lack of any better to place to write up some detail surrounding
-;;; layout creation for structure types, I'm putting here.
-;;; When you issue a DEFSTRUCT at the REPL, there are *three* instances
-;;; of LAYOUT makde for the new structure.
-;;; 1) The first is one associated with a temporary instance of
-;;; structure-classoid used in parsing the DEFSTRUCT form so that
-;;; we don't signal an UNKNOWN-TYPE condition for something like:
-;;;   (defstruct chain (next nil :type (or null chain)).
-;;; The temporary classoid is garbage immediately after parsing
-;;; and is never installed.
-;;; 2) The next is the actual LAYOUT that ends up being registered.
-;;; 3) The third is a layout created when setting the "compiler layout"
-;;; which contains copies of the length/depthoid/inherits etc
-;;; that we compare against the isntalled one to make sure they match.
-;;; The third one also gets thrown away.
-#-sb-xc-host
+#+metaspace
 (progn
-(defun make-layout (clos-hash classoid
-                    &key (depthoid -1) (length 0) (flags 0)
-                         (inherits #())
-                         (info nil)
-                         (bitmap (if info (dd-bitmap info) 0))
-                         (invalid :uninitialized))
-  (let* ((fixed-words (type-dd-length layout))
-         (bitmap-words (ceiling (1+ (integer-length bitmap)) sb-vm:n-word-bits))
-         (nwords (+ fixed-words bitmap-words))
-         (layout (truly-the layout
-                 #+immobile-space
-                 (sb-vm::alloc-immobile-fixedobj
-                  (1+ nwords)
-                  (logior (ash nwords sb-vm:instance-length-shift)
-                          sb-vm:instance-widetag))
-                 #-immobile-space (%make-instance nwords))))
-    (setf (%instance-layout layout) #.(find-layout 'layout))
-    (setf (layout-flags layout) #+64-bit (pack-layout-flags depthoid length flags)
-                                #-64-bit flags)
-    (setf (layout-clos-hash layout) clos-hash
-          (layout-classoid layout) classoid
-          (layout-invalid layout) invalid)
-    #-64-bit (setf (layout-depthoid layout) depthoid (layout-length layout) length)
-    (setf (layout-info layout) info)
-    (setf (layout-slot-list layout) nil
-          (layout-slot-table layout) #(1 nil))
-    (set-layout-inherits layout inherits)
-    #-sb-xc-host
-    (dotimes (i bitmap-words)
-      (setf (%raw-instance-ref/word layout (+ fixed-words i))
-            (ldb (byte sb-vm:n-word-bits (* i sb-vm:n-word-bits)) bitmap)))
-    (setf (layout-invalid layout) invalid)
-    layout))
-(declaim (inline layout-bitmap-words))
-(defun layout-bitmap-words (layout)
-  (declare (layout layout))
-  (- (%instance-length layout) (type-dd-length layout)))
-(defun layout-bitmap (layout)
-  (acond ((layout-info layout) (dd-bitmap it))
-         ;; Instances lacking DD-INFO are CLOS objects, which can't generally have
-         ;; raw slots, except that funcallable-instances may have 2 raw slots -
-         ;; the trampoline and the layout. The trampoline can have a tag, depending
-         ;; on the platform, and the layout is tagged but a special case.
-         ;; In any event, the bitmap is always 1 word.
-         (t (%raw-instance-ref/signed-word layout (type-dd-length layout))))))
+;;; Separating the pointer and nonpointer slots of LAYOUT satisfies the
+;;; requirement of a certain garbage collector (WIP), but also opens up the
+;;; possibility of absorbing at least one of the vectorish slots into the
+;;; WRAPPER since it could be reallocated with variable size.
+;;; The SLOT-TABLE slot might be a good candidate for trailing tagged slots.
+(sb-xc:defstruct (wrapper (:copier nil) (:constructor %make-wrapper))
+  ;; !!! The FRIEND slot in WRAPPER *MUST* BE FIRST !!! (Wired-in assumption in genesis)
+  (friend nil :type sb-vm:layout)
+  (clos-hash (missing-arg) :type (and fixnum unsigned-byte)) ; redundant
+  (classoid (missing-arg) :type classoid)
+  (inherits #() :type simple-vector)
+  (equalp-impl #'equalp-err :type (sfunction (t t) boolean) :read-only t)
+  (slot-table #(1 nil) :type simple-vector)
+  (%info nil :type (or list defstruct-description))
+  (invalid :uninitialized :type (or cons (member nil t :uninitialized))))
+;;; See #-metaspace structure definition for remarks about each slot.
+;;; LAYOUT points to WRAPPER and vice-versa.
+;;; The most common LAYOUT is 8 words.
+;;; Needing >64 words would be quite unusual - the layout would have
+;;; an enormous depthoid or bitmap or both.
+(sb-xc:defstruct (sb-vm:layout (:copier nil)
+                         ;; Parsing DEFSTRUCT uses a temporary layout
+                         (:constructor %make-temporary-layout (friend clos-hash)))
+  ;; !!! The FRIEND slot in LAYOUT *MUST* BE FIRST !!!
+  (friend nil :type wrapper)
+  (clos-hash (missing-arg) :type (and fixnum unsigned-byte))
+  (flags 0 :type (signed-byte #.sb-vm:n-word-bits))
+  (id-word0 0 :type word)
+  (id-word1 0 :type word)
+  (id-word2 0 :type word)
+  ;; There are zero or more raw words if a type needs to store additional layout-ids,
+  ;; and there are one or more raw words for the GC bitmap.
+  )
+(declaim (freeze-type wrapper sb-vm:layout)))
 
 ;;; The cross-compiler representation of a LAYOUT omits several things:
 ;;;   * BITMAP - obtainable via (DD-BITMAP (LAYOUT-INFO layout)).
 ;;;     GC wants it in the layout to avoid double indirection.
 ;;;   * EQUALP-TESTS - needed only for the target's implementation of EQUALP.
 ;;;   * SLOT-TABLE, and SLOT-LIST - used only by the CLOS implementation.
-;;;   * ANCESTOR_N are optimizations for TYPEP.
+;;;   * ID-WORDn are optimizations for TYPEP.
 ;;; So none of those really make sense on the host.
 ;;; Also, we eschew the packed representation of length+depthoid+flags.
-;;; FLAGS are not even strictly necessary, since they are for optimizing
-;;; various type checks.
+;;; FLAGS are computed on demand, and not stored.
 #+sb-xc-host
 (progn
-  (defstruct (layout (:include structure!object)
-                     (:constructor host-make-layout
-                                   (clos-hash classoid &key info depthoid inherits
-                                                       length flags invalid)))
+  (defstruct (wrapper (:include structure!object)
+                      (:constructor host-make-wrapper
+                         (id clos-hash classoid
+                          &key ((:info %info)) depthoid inherits length invalid
+                          #+metaspace friend)))
+    #+metaspace (friend)
+    (id nil :type (or null fixnum))
     ;; Cross-compiler-only translation from slot index to symbol naming
     ;; the accessor to call. (Since access by position is not a thing)
     (index->accessor-map #() :type simple-vector)
@@ -365,63 +288,123 @@
     ;; But there's no harm in storing it.
     (clos-hash nil :type (and sb-xc:fixnum unsigned-byte))
     (classoid nil :type classoid)
-    (flags 0 :type word)
     (invalid :uninitialized :type (or cons (member nil t :uninitialized)))
     (inherits #() :type simple-vector)
     (depthoid -1 :type layout-depthoid)
     (length 0 :type layout-length)
-    (info nil :type (or null defstruct-description)))
-  (defun make-layout (&rest args)
-    (let ((args (copy-list args)))
-      (remf args :bitmap)
-      (apply #'host-make-layout args)))
-  (defun make-temporary-layout (clos-hash classoid inherits)
-    (host-make-layout clos-hash classoid :inherits inherits :invalid nil))
-  (defun layout-bitmap (layout)
-    (if (layout-info layout) (dd-bitmap (layout-info layout)) +layout-all-tagged+)))
+    (%info nil :type (or null defstruct-description)))
+  #+metaspace (defstruct (sb-vm:layout (:include structure!object)
+                                       (:constructor %make-layout))
+                friend)
+  (defun make-temporary-wrapper (clos-hash classoid inherits)
+    (host-make-wrapper nil clos-hash classoid :inherits inherits :invalid nil))
+  (defun wrapper-flags (wrapper)
+    (declare (type wrapper wrapper))
+    (let ((mapping `((structure-object  ,+structure-layout-flag+)
+                     (standard-object   ,+pcl-object-layout-flag+)
+                     (pathname          ,+pathname-layout-flag+)
+                     (condition         ,+condition-layout-flag+)
+                     (file-stream       ,+file-stream-layout-flag+)
+                     (string-stream     ,+string-stream-layout-flag+)
+                     (stream            ,+stream-layout-flag+)
+                     (sequence          ,+sequence-layout-flag+)))
+          (flags 0))
+      (dolist (x (cons wrapper (coerce (wrapper-inherits wrapper) 'list)) flags)
+        (let ((cell (assoc (wrapper-classoid-name x) mapping)))
+          (when cell (setq flags (logior flags (second cell))))))))
+  (defun wrapper-bitmap (wrapper)
+    (acond ((wrapper-%info wrapper) (dd-bitmap it)) (t +layout-all-tagged+))))
 
-(defmacro sb-c::layout-nth-ancestor-slot (n)
-  `(case ,n
-     (2 'layout-ancestor_2)
-     (3 'layout-ancestor_3)
-     (4 'layout-ancestor_4)
-     (5 'layout-ancestor_5)))
+(defun equalp-err (a b)
+  (bug "EQUALP ~S ~S" a b))
 
-#+(and (not sb-xc-host) 64-bit)
-;;; LAYOUT-DEPTHOID gets a vop and a stub
-(defmacro layout-length (layout) ; SETFable
-  `(ldb (byte 16 16) (layout-flags ,layout)))
+(defmacro name->dd (name)
+  ;; This wants to be a toplevel macrolet but can't be, because the body of
+  ;; the macro (which is run in the host) wouldn't see NAME->DD as a macro
+  ;; when expanding for the target. And it can't be a toplevel FLET because
+  ;; that would demote the two using macros from toplevel.
+  `(find-defstruct-description (cond #-metaspace ((eq ,name 'sb-vm:layout) 'wrapper)
+                                     (t ,name))))
+(defmacro type-dd-length (type-name) (dd-length (name->dd type-name)))
+(defmacro get-dsd-index (type-name slot-name)
+  (declare (notinline dsd-index)) ; avoid later inlining failure style-warning
+  (dsd-index (find slot-name (dd-slots (name->dd type-name)) :key #'dsd-name)))
 
-(defconstant layout-flags-mask #xffff) ; "strictly flags" bits from the packed field
+;;; Applicable only if bit-packed (for 64-bit architectures)
+(defmacro pack-layout-flags (depthoid length flags)
+  `(logior (ash ,depthoid (+ 32 sb-vm:n-fixnum-tag-bits)) (ash ,length 16) ,flags))
 
-;;; Abstract out the differences between {32-bit,64-bit} target and XC layouts.
-;;; FLAGS can't change once assigned.
-(defmacro assign-layout-slots (layout &key depthoid length
-                                           (flags `(layout-flags ,layout)) flagsp)
-  (let ((invalidate-p (and (eql depthoid -1) (not length) (not flagsp))))
-    #+(and 64-bit (not xc-host)) ; packed slot
-    `(setf (layout-flags ,layout)
-           ,(if invalidate-p
-                `(logior (ash -1 (+ 32 sb-vm:n-fixnum-tag-bits))
-                         (ldb (byte 32 0) (layout-flags ,layout)))
-                `(pack-layout-flags ,depthoid ,length
-                                    (logand ,flags layout-flags-mask))))
-    #+(or (not 64-bit) sb-xc-host) ; ordinary slot
-    (if invalidate-p
-        `(setf (layout-depthoid ,layout) -1)
-        `(setf (layout-depthoid ,layout) ,depthoid
-               (layout-length ,layout) ,length
-               (layout-flags ,layout) ,flags))))
+(defconstant layout-id-vector-fixed-capacity 7)
+(defmacro calculate-extra-id-words (depthoid)
+  ;; There are 1 or 2 ids per word depending on n-word-bytes.
+  ;; We can always store IDs at depthoids 2,3,4,5,6,7,
+  ;; so depthoid less than or equal to 7 needs no extra words.
+  ;; 0 and 1 for T and STRUCTURE-OBJECT respectively are not stored.
+  `(ceiling (max 0 (- ,depthoid ,layout-id-vector-fixed-capacity))
+            ,(/ sb-vm:n-word-bytes 4)))
+
+(declaim (inline wrapper-info wrapper-dd))
+(defun wrapper-info (wrapper)
+  (let ((info (wrapper-%info wrapper))) (unless (listp info) info)))
+(defun (setf wrapper-info) (newval wrapper)
+  ;; The current value must be nil or a defstruct-description,
+  ;; otherwise we'd clobber a non-nil slot list.
+  (aver (not (consp (wrapper-%info wrapper))))
+  (setf (wrapper-%info wrapper) newval))
+;; Use WRAPPER-DD to read WRAPPER-INFO and assert that it is non-nil.
+(defun wrapper-dd (wrapper)
+  (the defstruct-description (wrapper-%info wrapper)))
+
+;;; See the pictures above DD-BITMAP in src/code/defstruct for the details.
+(defconstant standard-gf-primitive-obj-layout-bitmap
+  #+compact-instance-header  6
+  #-compact-instance-header -4)
+
+#-sb-xc-host
+(progn
+(declaim (inline bitmap-nwords bitmap-all-taggedp))
+(defun bitmap-nwords (layout)
+  (declare (sb-vm:layout layout))
+  (- (%instance-length layout) (type-dd-length sb-vm:layout)))
+
+(defun bitmap-all-taggedp (layout)
+  ;; All bitmaps have at least 1 word; read that first.
+  (and (= (%raw-instance-ref/signed-word layout (type-dd-length sb-vm:layout))
+          +layout-all-tagged+)
+       ;; Then check that there are no additional words.
+       (= (%instance-length layout) (1+ (type-dd-length sb-vm:layout)))))
+
+#+metaspace ; If metaspace, then WRAPPER has no flags; they're in the LAYOUT.
+(defmacro wrapper-flags (x) `(layout-flags (wrapper-friend ,x)))
+
+;; 32-bit has the depthoid as a slot, 64-bit has it is part of FLAGS which are
+;; in the LAYOUT
+#+64-bit
+(defun wrapper-depthoid (wrapper) (layout-depthoid (wrapper-friend wrapper)))
+
+(defun wrapper-bitmap (wrapper)
+  (declare (type wrapper wrapper))
+  (acond ((wrapper-info wrapper) (dd-bitmap it))
+         ;; Instances lacking DD-INFO are CLOS objects, which can't generally have
+         ;; raw slots, except that funcallable-instances may have 2 raw slots -
+         ;; the trampoline and the layout. The trampoline can have a tag, depending
+         ;; on the platform, and the layout is tagged but a special case.
+         ;; In any event, the bitmap is always 1 word, and there are no "extra ID"
+         ;; words preceding it.
+         (t (the fixnum
+                 (%raw-instance-ref/signed-word (wrapper-friend wrapper)
+                                                (type-dd-length sb-vm:layout))))))
+#+64-bit
+(defmacro wrapper-length (wrapper) ; SETFable
+  `(ldb (byte 16 16) (layout-flags (wrapper-friend ,wrapper))))
+
+) ; end PROGN
 
 ;;; True of STANDARD-OBJECT, which include generic functions.
-;;; This one includes any class that mixes in STANDARD-OBJECT.
 (declaim (inline layout-for-pcl-obj-p))
-(defun layout-for-pcl-obj-p (x)
-  (logtest (layout-flags x) +pcl-object-layout-flag+))
-
-(declaim (inline sb-fasl:dumpable-layout-p))
-(defun sb-fasl:dumpable-layout-p (x)
-  (and (typep x 'layout) (not (layout-for-pcl-obj-p x))))
+(defun layout-for-pcl-obj-p (wrapper)
+  (declare (type wrapper wrapper))
+  (logtest (wrapper-flags wrapper) +pcl-object-layout-flag+))
 
 ;;; The CLASSOID structure is a supertype of all classoid types.  A
 ;;; CLASSOID is also a CTYPE structure as recognized by the type
@@ -447,8 +430,8 @@
              #-sb-xc-host (:pure nil))
   ;; the value to be returned by CLASSOID-NAME.
   (name nil :type symbol)
-  ;; the current layout for this class, or NIL if none assigned yet
-  (layout nil :type (or layout null))
+  ;; the current WRAPPER for this class, or NIL if none assigned yet
+  (wrapper nil :type (or null wrapper))
   ;; How sure are we that this class won't be redefined?
   ;;   :READ-ONLY = We are committed to not changing the effective
   ;;                slots or superclasses.
@@ -465,14 +448,16 @@
   ;; otherwise, it's an EQ hash-table mapping CLASSOID objects to the
   ;; subclass layout that was in effect at the time the subclass was
   ;; created.
-  (subclasses nil :type (or null hash-table))
+  ;; Initially an alist, and changed to a hash-table at some threshold.
+  (subclasses nil :type (or list hash-table))
+  (%lock nil) ; install it just-in-time, similar to hash-table-lock
   ;; the PCL class (= CL:CLASS, but with a view to future flexibility
   ;; we don't just call it the CLASS slot) object for this class, or
   ;; NIL if none assigned yet
   (pcl-class nil))
 
-(defun layout-classoid-name (x)
-  (classoid-name (layout-classoid x)))
+(defun wrapper-classoid-name (x)
+  (classoid-name (wrapper-classoid x)))
 
 ;;;; object types to represent classes
 
@@ -494,13 +479,17 @@
 ;;; This translation is done when type specifiers are parsed. Type
 ;;; system operations (union, subtypep, etc.) should never encounter
 ;;; translated classes, only their translation.
-(def!struct (built-in-classoid (:include classoid)
-                               (:copier nil)
-                               (:constructor !make-built-in-classoid))
+(sb-xc:defstruct (built-in-classoid (:include classoid) (:copier nil)
+                                    (:constructor !make-built-in-classoid))
   ;; the type we translate to on parsing. If NIL, then this class
-  ;; stands on its own; or it can be set to :INITIALIZING for a period
-  ;; during cold-load.
-  (translation nil :type (or ctype (member nil :initializing))))
+  ;; stands on its own
+  (translation nil :type (or null ctype) :read-only t)
+  (predicate nil :type (sfunction (t) boolean) :read-only t))
+#+sb-xc-host
+(defstruct (built-in-classoid (:include classoid) (:copier nil)
+                              (:constructor !make-built-in-classoid))
+  ;; until bootstrap of all CTYPEs, store a dummy value distinct from NIL
+  (translation nil :type (or null ctype (member :initializing))))
 
 (def!struct (condition-classoid (:include classoid)
                                 (:copier nil)
@@ -530,6 +519,15 @@
   ;; environment of MAKE-CONDITION.
   (hairy-slots nil :type list))
 
+;;; STRUCTURE-CLASSOID represents what we need to know about structure
+;;; classes. Non-structure "typed" defstructs are a special case, and
+;;; don't have a corresponding class.
+(def!struct (structure-classoid
+             (:include classoid)
+             (:copier nil)
+             (:constructor make-structure-classoid
+                           (&key name &aux (%bits (pack-ctype-bits classoid name))))))
+
 ;;;; classoid namespace
 
 ;;; We use an indirection to allow forward referencing of class
@@ -552,26 +550,6 @@
     (declare (ignore env))
     `(find-classoid-cell ',(classoid-cell-name self) :create t)))
 
-(defun find-classoid-cell (name &key create)
-  (let ((real-name (uncross name)))
-    (cond ((info :type :classoid-cell real-name))
-          (create
-           (get-info-value-initializing :type :classoid-cell real-name
-                                        (make-classoid-cell real-name))))))
-
-;;; Return the classoid with the specified NAME. If ERRORP is false,
-;;; then NIL is returned when no such class exists.
-(defun find-classoid (name &optional (errorp t))
-  (declare (type symbol name))
-  (let ((cell (find-classoid-cell name)))
-    (cond ((and cell (classoid-cell-classoid cell)))
-          (errorp
-           (error 'simple-type-error
-                  :datum nil
-                  :expected-type 'class
-                  :format-control "Class not yet defined: ~S"
-                  :format-arguments (list name))))))
-
 ;;;; PCL stuff
 
 ;;; the CLASSOID that we use to represent type information for
@@ -593,49 +571,3 @@
 
 (declaim (freeze-type built-in-classoid condition-classoid
                       standard-classoid static-classoid))
-
-(in-package "SB-C")
-
-;;; layout for this type being used by the compiler
-(define-info-type (:type :compiler-layout)
-  :type-spec (or layout null)
-  :default (lambda (name)
-             (awhen (find-classoid name nil) (classoid-layout it))))
-
-(eval-when (#-sb-xc :compile-toplevel :load-toplevel :execute)
-(defun ftype-from-fdefn (name)
-  (declare (ignorable name))
-  ;; Again [as in (DEFINE-INFO-TYPE (:FUNCTION :TYPE) ...)] it's
-  ;; not clear how to generalize the FBOUNDP expression to the
-  ;; cross-compiler. -- WHN 19990330
-  #+sb-xc-host
-  (specifier-type 'function)
-  #-sb-xc-host
-  (let* ((fdefn (sb-kernel::find-fdefn name))
-         (fun (and fdefn (fdefn-fun fdefn))))
-    (if fun
-        (handler-bind ((style-warning #'muffle-warning))
-          (specifier-type (sb-impl::%fun-type fun)))
-        (specifier-type 'function)))))
-
-;;; The parsed or unparsed type for this function, or the symbol :GENERIC-FUNCTION.
-;;; Ordinarily a parsed type is stored. Only if the parsed type contains
-;;; an unknown type will the original specifier be stored; we attempt to reparse
-;;; on each lookup, in the hope that the type becomes known at some point.
-;;; If :GENERIC-FUNCTION, the info is recomputed from methods at the time of lookup
-;;; and stored back. Method redefinition resets the value to :GENERIC-FUNCTION.
-(define-info-type (:function :type)
-  :type-spec (or ctype (cons (eql function)) (member :generic-function))
-  :default #'ftype-from-fdefn)
-
-(defun summarize-layouts ()
-  (let ((prev -1))
-    (dolist (layout (sort (loop for v being each hash-value
-                                of (classoid-subclasses (find-classoid 't))
-                                collect v)
-                          #'< :key #'sb-kernel:layout-flags))
-      (let ((flags (sb-kernel:layout-flags layout)))
-        (unless (= flags prev)
-          (format t "Layout flags = ~d~%" flags)
-          (setq prev flags)))
-      (format t "  ~a~%" layout))))
