@@ -11,8 +11,6 @@
 
 (in-package "SB-VM")
 
-(eval-when (:compile-toplevel :load-toplevel :execute)
-  (export 'sb-sys::get-page-size "SB-SYS"))
 
 ;;;; type format database
 
@@ -34,7 +32,7 @@
         (room-info-name info)))
 
 (defconstant tiny-boxed-size-mask #xFF)
-(defun !compute-room-infos ()
+(defun compute-room-infos ()
   (let ((infos (make-array 256 :initial-element nil))
         (default-size-mask (mask-field (byte 23 0) -1)))
     (dolist (obj *primitive-objects*)
@@ -49,11 +47,11 @@
                  (symbol (make-room-info tiny-boxed-size-mask name :other))
                  (t      (make-room-info default-size-mask name :other)))))))
 
-    (let ((info (make-room-info default-size-mask 'array-header :other)))
+    (let ((info (make-room-info tiny-boxed-size-mask 'array-header :other)))
       (dolist (code (list #+sb-unicode complex-character-string-widetag
                           complex-base-string-widetag simple-array-widetag
                           complex-bit-vector-widetag complex-vector-widetag
-                          complex-array-widetag complex-vector-nil-widetag))
+                          complex-array-widetag))
         (setf (svref infos code) info)))
 
     (setf (svref infos bignum-widetag)
@@ -73,6 +71,7 @@
         (when (saetp-specifier saetp) ;; SIMPLE-ARRAY-NIL is a special case.
           (setf (svref infos (saetp-typecode saetp)) saetp))))
 
+    ;; This one is here for completeness only- LENGTH does not imply size.
     (setf (svref infos simple-array-nil-widetag)
           (make-room-info 0 'simple-array-nil :other))
 
@@ -112,7 +111,8 @@
 
     infos))
 
-(define-load-time-global *room-info* (!compute-room-infos))
+(define-load-time-global *room-info* (compute-room-infos))
+(declaim (type (simple-vector 256) *room-info*))
 
 (defconstant-eqx +heap-spaces+
   '((:dynamic   "Dynamic space"   dynamic-usage)
@@ -207,52 +207,47 @@
 ;;; But see WITH-PINNED-OBJECT-ITERATOR in 'target-hash-table' which offers
 ;;; a way to mitigate the need to bind/unbind a special var when doing a massive
 ;;; numbers of WITH-PINNED-OBJECT operations.
-;;; FIXME: export this from SB-EXT
 (defun primitive-object-size (object)
   "Return number of bytes of heap or stack directly consumed by OBJECT"
-  (if (is-lisp-pointer (get-lisp-obj-address object))
-      (let ((words
-              (typecase object
-                (cons 2)
-                (instance (1+ (instance-length object)))
-                (function
-                 (when (= (fun-subtype object) simple-fun-widetag)
-                   (return-from primitive-object-size
-                     (primitive-object-size (fun-code-header object))))
-                 (1+ (get-closure-length object)))
-                (null sizeof-nil-in-words)
-                (code-component
-                 (return-from primitive-object-size (code-object-size object)))
-                (fdefn 4) ; no length stored in the header
-                ;; Anything else is an OTHER pointer.
-                (t
-                 (let ((room-info
-                         (aref *room-info* (%other-pointer-widetag object))))
-                   (if (arrayp object)
-                       (cond ((array-header-p object)
-                              (+ array-dimensions-offset (array-rank object)))
-                             ((simple-array-nil-p object) 2)
-                             (t
-                              (let* ((length (+ (length object)
-                                                (saetp-n-pad-elements room-info)))
-                                     (n-bits (saetp-n-bits room-info))
-                                     (alignment-pad (floor 7 n-bits))
-                                     ;; This math confuses me. So there's a self-test
-                                     ;; against the C code which is known good.
-                                     (n-data-octets (if (>= n-bits 8)
-                                                        (* length (ash n-bits -3))
-                                                        (ash (* (+ length alignment-pad)
-                                                                n-bits)
-                                                             -3))))
-                                (+ (ceiling n-data-octets n-word-bytes) ; N data words
-                                   vector-data-offset))))
-                       ;; Other things (symbol, value-cell, etc)
-                       ;; don't have a sizer, so use GET-HEADER-DATA
-                       (1+ (logand (get-header-data object)
-                                   (room-info-mask room-info)))))))))
-        (* (align-up words 2) n-word-bytes))
-      0))
+  (unless (is-lisp-pointer (get-lisp-obj-address object))
+    (return-from primitive-object-size 0))
+  (let ((words
+         ;; This should pick off the most frequently-occurring things first.
+         ;; CONS and INSTANCE of course top the list.
+         (typecase object
+           (cons 2)
+           (instance (1+ (instance-length object)))
+           (function
+            (when (= (%fun-pointer-widetag object) simple-fun-widetag)
+              (return-from primitive-object-size
+                (code-object-size (fun-code-header (truly-the simple-fun object)))))
+            (1+ (get-closure-length object)))
+           (code-component
+            (return-from primitive-object-size (code-object-size object)))
+           ;; Most everything else is an OTHER pointer, except for NIL.
+           ;; Arrays (especially strings and simple-vector) and symbols tend to be the
+           ;; next-most-common heap object, for which we need the ROOM-INFO.
+           (t
+            (let ((room-info (aref *room-info* (%other-pointer-widetag object))))
+              (cond ((typep room-info 'specialized-array-element-type-properties)
+                     (let ((n-data-octets (vector-n-data-octets object room-info)))
+                       (+ (ceiling n-data-octets n-word-bytes) ; N data words
+                          vector-data-offset)))
+                    (t
+                     (typecase object
+                       (fdefn 4) ; constant length not stored in the header
+                       ((satisfies array-header-p)
+                        (+ array-dimensions-offset (array-rank object)))
+                       ((simple-array nil (*)) 2) ; no payload
+                       (null sizeof-nil-in-words)
+                       (t
+                        ;; GET-HEADER-DATA works for everything that's left
+                        (1+ (logand (get-header-data object)
+                                    (room-info-mask room-info))))))))))))
+        (* (align-up words 2) n-word-bytes)))
 
+;;; Macros not needed after this file (and avoids a redefinition warning this way)
+(eval-when (:compile-toplevel)
 (defmacro widetag@baseptr (sap)
   #+big-endian `(sap-ref-8 ,sap ,(1- n-word-bytes))
   #+little-endian `(sap-ref-8 ,sap 0))
@@ -261,7 +256,7 @@
   `(%make-lisp-obj
     (logior (sap-int ,sap)
             (logand (deref (extern-alien "widetag_lowtag" (array char 256)) ,widetag)
-                    lowtag-mask))))
+                    lowtag-mask)))))
 
 ;;; This uses the funny fixnum representation of ADDRESS. I'd like to change this
 ;;; to take a SAP but god forbid people are using it?
@@ -300,6 +295,13 @@
        (setq start (sap+ start size))))
     (when strict-bound
      ;; If START is not eq to END, then we have blown past our endpoint.
+      #+sb-devel
+     (unless (sap= start end)
+       ;; don't make things go more wrong than they already are.
+       (alien-funcall (extern-alien "printf" (function void system-area-pointer))
+                      (vector-sap #.(format nil "map-objects-in-range failure~%")))
+       (alien-funcall (extern-alien "ldb_monitor" (function void))))
+     #-sb-devel
      (aver (sap= start end)))))
 
 ;;; Test the sizing function ASAP, because if it's broken, then MAP-ALLOCATED-OBJECTS
@@ -332,7 +334,6 @@
               ;; of word size. This is fine for 32-bit address space,
               ;; but if 64-bit then we have to scale the value. Additionally
               ;; there is a fallback for when even the scaled value is too big.
-              ;; (None of this matters to Lisp code for the most part)
               (start #+64-bit (unsigned 32) #-64-bit signed)
               ;; On platforms with small enough GC pages, this field
               ;; will be a short. On platforms with larger ones, it'll
@@ -431,41 +432,59 @@ We could try a few things to mitigate this:
   ;; You can't specify :ALL and also a list of spaces. Check that up front.
   (do-rest-arg ((space) spaces) (the spaces space))
   (flet ((do-1-space (space)
-    (ecase space
-      (:static
-       ;; Static space starts with NIL, which requires special
-       ;; handling, as the header and alignment are slightly off.
-       (multiple-value-bind (start end) (%space-bounds space)
-         (declare (ignore start))
-         (funcall fun nil symbol-widetag (* sizeof-nil-in-words n-word-bytes))
-         (let ((start (+ (logandc2 sb-vm:nil-value sb-vm:lowtag-mask)
-                         (ash (- sizeof-nil-in-words 2) sb-vm:word-shift))))
-           (map-objects-in-range fun
-                                 (ash start (- sb-vm:n-fixnum-tag-bits))
-                                 end))))
+           (ecase space
+            (:static
+             ;; Static space starts with NIL, which requires special
+             ;; handling, as the header and alignment are slightly off.
+             (multiple-value-bind (start end) (%space-bounds space)
+               (declare (ignore start))
+               (funcall fun nil symbol-widetag (* sizeof-nil-in-words n-word-bytes))
+               (let ((start (+ (logandc2 sb-vm:nil-value sb-vm:lowtag-mask)
+                               (ash (- sizeof-nil-in-words 2) sb-vm:word-shift))))
+                 (map-objects-in-range fun
+                                       (ash start (- sb-vm:n-fixnum-tag-bits))
+                                       end))))
 
-      ((:read-only #-gencgc :dynamic)
-       ;; Read-only space (and dynamic space on cheneygc) is a block
-       ;; of contiguous allocations.
-       (multiple-value-bind (start end) (%space-bounds space)
-         (map-objects-in-range fun start end)))
-      #+immobile-space
-      (:immobile
-       ;; Filter out filler objects. These either look like cons cells
-       ;; in fixedobj subspace, or code without enough header words
-       ;; in varyobj subspace. (cf 'filler_obj_p' in gc-internal.h)
-       (dx-flet ((filter (obj type size)
-                   (unless (= type list-pointer-lowtag)
-                     (funcall fun obj type size))))
-         (map-immobile-objects #'filter :fixed))
-       (dx-flet ((filter (obj type size)
-                   (unless (and (code-component-p obj)
-                                (code-obj-is-filler-p obj))
-                     (funcall fun obj type size))))
-         (map-immobile-objects #'filter :variable)))
+            ((:read-only #-gencgc :dynamic)
+             ;; Read-only space (and dynamic space on cheneygc) is a block
+             ;; of contiguous allocations.
+             (multiple-value-bind (start end) (%space-bounds space)
+                                  (map-objects-in-range fun start end)))
+            #+immobile-space
+            (:immobile
+             ;; Filter out filler objects. These either look like cons cells
+             ;; in fixedobj subspace, or code without enough header words
+             ;; in varyobj subspace. (cf 'filler_obj_p' in gc-internal.h)
+             (dx-flet ((filter (obj type size)
+                         (unless (= type list-pointer-lowtag)
+                           (funcall fun obj type size))))
+               (map-immobile-objects #'filter :fixed))
+             (dx-flet ((filter (obj type size)
+                         (unless (and (code-component-p obj)
+                                      (code-obj-is-filler-p obj))
+                           (funcall fun obj type size))))
+               (map-immobile-objects #'filter :variable))))))
+    (do-rest-arg ((space) spaces)
+      (if (eq space :dynamic)
+          (without-gcing #+cheneygc (do-1-space space)
+                         #+gencgc (walk-dynamic-space fun #b1111111 0 0))
+          (do-1-space space)))))
 
-      #+gencgc
-      (:dynamic
+;;; Using the mask bits you can make many different match conditions resulting
+;;; from a product of {boxed,unboxed,code,any} x {large,non-large,both}
+;;; e.g. mask = #b10011" constraint = "#b10010"
+;;; matches "large & (unboxed | code)"
+;;;
+;;; I think, when iterating over only code, that if we grab the code_allocator_lock
+;;; and free_pages_lock, that this can be made reliable (both crash-free and
+;;; guaranteed to visit all chosen objects) despite other threads running.
+;;; As things are it is only "maybe" reliable, regardless of the parameters.
+#+gencgc
+(defun walk-dynamic-space (fun generation-mask
+                               page-type-mask page-type-constraint)
+  (declare (function fun)
+           (type (unsigned-byte 7) generation-mask)
+           (type (unsigned-byte 5) page-type-mask page-type-constraint))
        ;; Dynamic space on gencgc requires walking the GC page tables
        ;; in order to determine what regions contain objects.
 
@@ -481,87 +500,57 @@ We could try a few things to mitigate this:
        ;; BYTES-USED is not GENCGC-CARD-BYTES or we reach
        ;; NEXT-FREE-PAGE.  We then MAP-OBJECTS-IN-RANGE if the range
        ;; is not empty, and proceed to the next page (unless we've hit
-       ;; NEXT-FREE-PAGE).  We happily take advantage of the fact that
-       ;; MAP-OBJECTS-IN-RANGE will simply return if passed two
-       ;; coincident pointers for the range.
+       ;; NEXT-FREE-PAGE).
 
        ;; FIXME: WITHOUT-GCING prevents a GC flip, but doesn't prevent
        ;; closing allocation regions and opening new ones.  This may
        ;; prove to be an issue with concurrent systems, or with
        ;; spectacularly poor timing for closing an allocation region
        ;; in a single-threaded system.
-       (close-current-gc-region)
-       (do ((initial-next-free-page next-free-page)
-            ;; This is a "funny" fixnum - essentially the bit cast of a pointer
-            (start (the fixnum (%make-lisp-obj (current-dynamic-space-start))))
-            (page-index 0 (1+ page-index))
-            ;; SPAN is a page count, for which an unsigned fixnum is adequate.
-            (span 0))
-           ;; We can safely iterate up to and including next_free_page
-           ;; even if next_free_page is the total number of pages in the space,
-           ;; because there is one extra page table entry as a sentinel.
-           ;; The extra page always has 0 bytes used, so we'll observe the end
-           ;; of a contiguous block without needing a termination clause to
-           ;; handle a final sequence of totally full pages that exactly abut
-           ;; the heap end. [Also note that for our purposes, contiguous blocks
-           ;; can span different GC generations and page types, whereas within
-           ;; GC, a page ends a block if the next differs in those aspects]
-           ((> page-index initial-next-free-page))
-         ;; The type constraint on PAGE-INDEX is probably too generous,
+  (close-current-gc-region)
+  (do ((initial-next-free-page next-free-page)
+       (base (int-sap (current-dynamic-space-start)))
+       (start-page 0)
+       (end-page 0)
+       (end-page-bytes-used 0))
+      ((> start-page initial-next-free-page))
+         ;; The type constraint on page indices is probably too generous,
          ;; but it does its job of producing efficient code.
-         (declare (type (integer 0 (#.(/ (ash 1 n-machine-word-bits) gencgc-card-bytes)))
-                        page-index)
-                  (type (and fixnum unsigned-byte) span))
-         (let ((page-bytes-used ; The low bit of bytes-used is the need-to-zero flag.
-                (logandc1 1 (slot (deref page-table page-index) 'bytes-used))))
-           (if (= page-bytes-used gencgc-card-bytes)
-               (incf span)
-               ;; RANGE-END forces funny increment of START by the extent in bytes,
-               ;; returning a negative fixum when the word's high bit flips.
-               ;; The proper way to add "funny" fixnums is via +-MODFX which for
-               ;; reasons unknown isn't defined on all backends.
-               (macrolet
-                   ((range-end (extra)
-                      `(truly-the fixnum
-                        (%make-lisp-obj
-                         (logand most-positive-word
-                                 ;; The first and third addends are known good
-                                 ;; The second needs help to avoid an expensive ASH
-                                 (+ (get-lisp-obj-address start)
-                                    (logand (* span gencgc-card-bytes) most-positive-word)
-                                    ,extra))))))
-                 (map-objects-in-range fun start (range-end page-bytes-used)
-                                       (< page-index initial-next-free-page))
-                 ;; In a certain rare situation, we must restart the next loop iteration
-                 ;; at exactly PAGE-INDEX instead of 1+ PAGE-INDEX.
-                 ;; Normally the contents of PAGE-INDEX are always included in the
-                 ;; current range. But what if it contributed 0 bytes to the range?
-                 ;;
-                 ;;     N : #x8000 bytes used
-                 ;;   N+1 : #x8000 bytes used
-                 ;;   N+2 : 0 bytes used        <- PAGE-INDEX now points here
-                 ;;   N+3 : initially 0 bytes, then gets some bytes
-                 ;;
-                 ;; We invoked the map function with page_address(N) for #x10000 bytes.
-                 ;; Suppose that function consed a partly unboxed object starting on
-                 ;; page N+2 extending to page N+3, and that NEXT-FREE-PAGE was greater
-                 ;; than both to begin with, so the termination condition isn't in play.
-                 ;; In the best case scenario, resuming the scan at page N+3 (mid-object)
-                 ;; would read valid widetags, but in the worst case scenario, we get
-                 ;; random bits. Page N+2 needs a fighting chance to be the start of
-                 ;; a range, so go back 1 page if the current page had zero bytes used
-                 ;; and the span exceeded 1. That is, always make forward progress.
-                 (setq start (range-end (cond ((and (zerop page-bytes-used) (plusp span))
-                                               (decf page-index)
-                                               0)
-                                              (t
-                                               gencgc-card-bytes)))
-                       span 0)))))))))
-
-  (do-rest-arg ((space) spaces)
-    (if (eq space :dynamic)
-        (without-gcing (do-1-space space))
-        (do-1-space space)))))
+    (declare (type (integer 0 (#.(/ (ash 1 n-machine-word-bits) gencgc-card-bytes)))
+                   start-page end-page))
+    (setq end-page start-page)
+    (loop (setq end-page-bytes-used
+                ;; The low bit of bytes-used is the need-to-zero flag.
+                (logandc1 1 (slot (deref page-table end-page) 'bytes-used)))
+          ;; See 'page_ends_contiguous_block_p' in gencgc.c
+          (when (or (< end-page-bytes-used gencgc-card-bytes)
+                    (= (slot (deref page-table (1+ end-page)) 'start) 0))
+            (return))
+          (incf end-page))
+    (let ((start (sap+ base (truly-the sb-vm:signed-word
+                                       (logand (* start-page gencgc-card-bytes)
+                                               most-positive-word))))
+          (end (sap+ base (truly-the sb-vm:signed-word
+                                     (logand (+ (* end-page gencgc-card-bytes)
+                                                end-page-bytes-used)
+                                             most-positive-word)))))
+      (when (sap> end start)
+        ;; The bits in the 5-bit flag field have fixed positions,
+        ;; but the position of the field itself depends on endianness.
+        (let ((flags (ldb #+little-endian (byte 5 0) #+big-endian (byte 5 3)
+                          (slot (deref page-table start-page) 'flags))))
+          ;; The GEN slot is declared as (SIGNED 8) which does not satisfy the
+          ;; type restriction on the first argument to LOGBITP.
+          ;; Masking it to 3 bits fixes that, and allows using the other 5 bits
+          ;; for something potentially.
+          (when (and (logbitp (logand (slot (deref page-table start-page) 'gen) 7)
+                              generation-mask)
+                     (= (logand flags page-type-mask) page-type-constraint))
+            (map-objects-in-range fun
+                                  (%make-lisp-obj (sap-int start))
+                                  (%make-lisp-obj (sap-int end))
+                                  (< start-page initial-next-free-page))))))
+    (setq start-page (1+ end-page))))
 
 ;; Start with a Lisp rendition of ensure_region_closed() on the active
 ;; thread's region, since users are often surprised to learn that a
@@ -790,9 +779,9 @@ We could try a few things to mitigate this:
                  (eql type funcallable-instance-widetag))
          (incf total-objects)
          (let* ((layout (if (eql type funcallable-instance-widetag)
-                            (%fun-layout obj)
-                            (%instance-layout obj)))
-                (classoid (layout-classoid layout))
+                            (%fun-wrapper obj)
+                            (%instance-wrapper obj)))
+                (classoid (wrapper-classoid layout))
                 (found (ensure-gethash classoid totals (cons 0 0)))
                 (size size))
            (declare (fixnum size))
@@ -847,11 +836,6 @@ We could try a few things to mitigate this:
 
 ;;;; PRINT-ALLOCATED-OBJECTS
 
-;;; This notion of page-size is completely arbitrary - it affects 2 things:
-;;; (1) how much output to print "per page" in print-allocated-objects
-;;; (2) sb-sprof deciding how many regions [sic] were made if #+cheneygc
-(defun get-page-size () sb-c:+backend-page-bytes+)
-
 ;;; This function is sheer madness.  You're better off using
 ;;; LIST-ALLOCATED-OBJECTS and then iterating over that, to avoid
 ;;; seeing all the junk created while doing this thing.
@@ -865,7 +849,7 @@ We could try a few things to mitigate this:
     (let* ((space-start (ash start n-fixnum-tag-bits))
            (space-end (ash end n-fixnum-tag-bits))
            (space-size (- space-end space-start))
-           (pagesize (get-page-size))
+           (pagesize sb-c:+backend-page-bytes+)
            (start (+ space-start (round (* space-size percent) 100)))
            (printed-conses (make-hash-table :test 'eq))
            (pages-so-far 0)
@@ -1067,12 +1051,9 @@ We could try a few things to mitigate this:
          ;; These two are in fact generally the most frequently occurring type.
          ,.(make-case 'cons `(car ,obj) `(cdr ,obj))
          ,.(make-case* 'instance
-            `(let ((.l. (%instance-layout ,obj)))
-               ;; Though we've bound %INSTANCE-LAYOUT to a variable,
-               ;; pass the form %INSTANCE-LAYOUT to functoid
-               ;; in case it wants to examine the form
+            `(progn
                (,functoid (%instance-layout ,obj) ,@more)
-               (do-instance-tagged-slot (.i. ,obj :layout .l. :pad nil)
+               (do-instance-tagged-slot (.i. ,obj nil)
                  (,functoid (%instance-ref ,obj .i.) ,@more))))
          (function
           (typecase ,obj
@@ -1084,23 +1065,39 @@ We could try a few things to mitigate this:
                   ;; if functoid is a macro that does nothing.
                   (,functoid .o. ,@more)))
             ,.(make-case* 'funcallable-instance
-               `(let ((.l. (%fun-layout ,obj)))
+               `(progn
                   ;; As for INSTANCE, allow the functoid to see the access form
                   (,functoid (%fun-layout ,obj) ,@more)
                   (,functoid (%funcallable-instance-fun ,obj) ,@more)
-                  (ecase (layout-bitmap .l.)
-                    (#.sb-kernel:+layout-all-tagged+
-                     (loop for .i. from instance-data-start ; exclude layout
+                  ;; Unfortunately for FUNCALLABLE-INSTANCEs, the relation
+                  ;; between layout bitmap indices and indices as given to
+                  ;; FUNCALLABLE-INSTANCE-INFO is not so obvious, and it's
+                  ;; both tricky and unnecessary to generalize iteration.
+                  ;; So just hardcode the few cases that exist.
+                  #+compact-instance-header
+                  (ecase (wrapper-bitmap (%fun-wrapper ,obj))
+                    (-1 ; external trampoline, all slots are tagged
+                     ;; In this case, the trampoline word is scanned, with no ill effect.
+                     (loop for .i. from 0
                            to (- (get-closure-length ,obj) funcallable-instance-info-offset)
                            do (,functoid (%funcallable-instance-info ,obj .i.) ,@more)))
-                    (#b0110
-                     ;; A pedantically correct kludge which shall remain unless need arises
-                     ;; for more general partially unboxed FINs.
-                     ;;  payload word 0 is raw (but looks like a fixnum, by design)
-                     ;;  word 1 is the fin-fun which we already accounted for above
-                     ;;  word 2 (info slot 0) is the only one that hasn't been processed.
-                     ;;  words 3 and 4 are raw but looks like fixnums by accident.
-                     (,functoid (%funcallable-instance-info ,obj 0) ,@more)))))
+                    (#b0110 ; internal trampoline, 2 raw slots, 1 tagged slot
+                     ;;   ^ ---- trampoline
+                     ;;  ^------ implementation function
+                     ;; ^------- (FUNCALLABLE-INSTANCE-INFO 0)
+                     ;; and the rest of the words are raw.
+                     (,functoid (%funcallable-instance-info ,obj 0) ,@more)))
+                  #-compact-instance-header
+                  (progn
+                    (aver (eql (wrapper-bitmap (%fun-wrapper ,obj)) -4))
+                    ;;            v ----trampoline
+                    ;; = #b1...1100
+                    ;;           ^----- layout
+                    ;;          ^------ implementation function
+                    ;;         ^------- (FUNCALLABLE-INSTANCE-INFO 0)
+                    (loop for .i. from 0
+                          to (- (get-closure-length ,obj) funcallable-instance-info-offset)
+                          do (,functoid (%funcallable-instance-info ,obj .i.) ,@more)))))
             .,(make-case 'function))) ; in case there was code provided for it
          (t
           ;; TODO: the generated code is pretty horrible. OTHER-POINTER-LOWTAG
@@ -1162,6 +1159,7 @@ We could try a few things to mitigate this:
 ;;; code-components are considered to reference their embedded
 ;;; simple-funs for this purpose; if THIS is a simple-fun, it is ignored.
 (defun references-p (this that)
+  (declare (optimize (sb-c::aref-trapping 0)))
   (macrolet ((test (x) `(when (eq ,x that) (go win))))
     (tagbody
        (do-referenced-object (this test)
@@ -1321,7 +1319,7 @@ We could try a few things to mitigate this:
 (defun show-immobile-spaces (which)
   (flet ((show (obj type size)
            (declare (ignore type size))
-           (let ((*print-pretty* nil))
+           (let ((*print-pretty* nil) (*print-length* 3))
              (format t "~x: ~s~%" (get-lisp-obj-address obj) obj))))
     (when (or (eq which :fixed) (eq which :both))
       (format t "Fixedobj space~%==============~%")
@@ -1453,6 +1451,103 @@ We could try a few things to mitigate this:
         (format t "~5d = ~s~%" n x)
         (setq prev n)))))
 
+#+gencgc
+(flet ((print-it (obj type size)
+         (declare (ignore type size))
+         (let ((*print-level* 2) (*print-lines* 1))
+           (format t "~x ~s~%" (get-lisp-obj-address obj) obj))))
+(defun print-all-code ()
+  (walk-dynamic-space #'print-it #x7f #b11 #b11))
+(defun print-large-code ()
+  (walk-dynamic-space #'print-it #x7f #b10011 #b10011))
+(defun print-large-unboxed ()
+  (walk-dynamic-space #'print-it #x7f #b10011 #b10010))
+;;; Use this only if you know what you're doing. It can fail because a page
+;;; that needs to continue onto the next page will cause the "overrun" check
+;;; to fail.
+(defun print-page-contents (page)
+  (let* ((start
+          (+ (current-dynamic-space-start) (* sb-vm:gencgc-card-bytes page)))
+         (end
+          (+ start sb-vm:gencgc-card-bytes)))
+    (map-objects-in-range #'print-it (%make-lisp-obj start) (%make-lisp-obj end)))))
+
+(defun map-code-objects (fun)
+  (dx-flet ((filter (obj type size)
+              (declare (ignore size))
+              (when (= type sb-vm:code-header-widetag)
+                (funcall fun obj)))
+            (nofilter (obj type size)
+              (declare (ignore type size))
+              (funcall fun obj)))
+    #+cheneygc (map-allocated-objects #'filter :all)
+    #+gencgc
+    (without-gcing
+      #+immobile-code
+      (map-objects-in-range #'nofilter
+                            (ash varyobj-space-start (- n-fixnum-tag-bits))
+                            (%make-lisp-obj (sap-int *varyobj-space-free-pointer*)))
+      (alien-funcall (extern-alien "close_code_region" (function void)))
+      (walk-dynamic-space #'nofilter
+                          #b1111111 ; all generations
+                          3 3))))
+
+(export 'code-from-serialno)
+(defun code-from-serialno (serial)
+  (dx-flet ((visit (obj)
+              (when (= (%code-serialno obj) serial)
+                (return-from code-from-serialno obj))))
+    (map-code-objects #'visit)))
+
+(defun show-all-layouts ()
+  (let ((l (sb-vm::list-allocated-objects :all :test #'sb-kernel::wrapper-p))
+        zero trailing-raw trailing-tagged vanilla)
+    (dolist (x l)
+      (let ((m (wrapper-bitmap x)))
+        (cond ((eql m +layout-all-tagged+) (push x vanilla))
+              ((eql m 0) (push x zero))
+              ((minusp m) (push x trailing-tagged))
+              (t (push x trailing-raw)))))
+    (flet ((legend (newline str list)
+             (when newline (terpri))
+             (let ((s (format nil str (length list))))
+               (format t "~A~%~A~%" s (make-string (length s) :initial-element #\-)))))
+      (when zero
+        (legend nil "Zero bitmap (~d):" zero)
+        (dolist (x zero) (format t "~a~%" (wrapper-classoid-name x))))
+      (when trailing-raw
+        (legend t "Trailing raw (~d):" trailing-raw)
+        (dolist (x trailing-raw)
+          (let ((m (wrapper-bitmap x)))
+            (format t "~30a 0...~v,'0b~%"
+                    (wrapper-classoid-name x)
+                    (acond ((wrapper-info x) (1+ (dd-length it))) (t 32))
+                    m))))
+      (when trailing-tagged
+        (legend t "Trailing tagged (~d):" trailing-tagged)
+        (dolist (x trailing-tagged)
+          (let ((m (wrapper-bitmap x)))
+            (format t "~30a 1...~b~%"
+                    (wrapper-classoid-name x)
+                    (acond ((wrapper-info x) (ldb (byte (dd-length it) 0) m))
+                           (t (ldb (byte 32 0) m)))))))
+      (legend t "Default: (~d) [not shown]" vanilla))))
+
+#+ubsan
+(defun find-poisoned-vectors (&aux result)
+  (dolist (v (sb-vm:list-allocated-objects :all :type sb-vm:simple-vector-widetag)
+             result)
+    (when (dotimes (i (length v))
+            (declare (optimize (sb-c::aref-trapping 0)))
+            (let ((val (svref v i)))
+              (when (= (get-lisp-obj-address val) no-tls-value-marker-widetag)
+                (return t))))
+      (push (make-weak-pointer v) result)
+      (let* ((origin (vector-extra-data v))
+             (code (sb-di::code-header-from-pc (ash origin -3)))
+             (*print-array* nil))
+        (format t "g~d ~a ~a~%" (sb-kernel:generation-of v) v code)))))
+
 (in-package "SB-C")
 ;;; As soon as practical in warm build it makes sense to add
 ;;; cold-allocation-point-fixups into the weak hash-table.
