@@ -14,34 +14,28 @@
 
 ;;; various environment inquiries
 
-;;; Compute the target's value of CL:*FEATURES* be deleting non-target features.
-;;; - :CONS-PROFILING   is used only to assign a compiler policy which persists
-;;;                     into the default baseline policy. It has no relevance post-build
-;;;                     in as much as policy can be changed later arbitrarily.
-;;; - :GCC-TLS          is not a Lisp feature- it's just freeloading off the means
-;;;                     by which additional #defines get into "genesis/config.h".
-;;; - :SB-AFTER-XC-CORE is merely a build option.
-;;; - :SB-XC            indicates the build phase and must not persist.
-(defparameter *features*
-   '#.(remove-if (lambda (x) (member x '(:cons-profiling :gcc-tls :sb-after-xc-core :sb-xc)))
-                 sb-xc:*features*)
+;;; This is a tentative list of target features; many are removed later.
+;;; :SB-XC is removed now, because it is plain wrong unless cross-compiling.
+(defparameter *features* '#.(remove :sb-xc sb-xc:*features*)
   "a list of symbols that describe features provided by the
    implementation")
+(defconstant !sbcl-architecture #.(sb-cold::target-platform-keyword))
 
 (defun machine-instance ()
   "Return a string giving the name of the local machine."
   #+win32 (sb-win32::get-computer-name)
   #-win32 (truly-the simple-string (sb-unix:unix-gethostname)))
 
-(declaim (type (or null string) *machine-version*))
-(defvar *machine-version*)
+(declaim (type (or null simple-string) *machine-version*))
+(declaim (global *machine-version*))
 
 (defun machine-version ()
   "Return a string describing the version of the computer hardware we
 are running on, or NIL if we can't find any useful information."
-  (unless (boundp '*machine-version*)
-    (setf *machine-version* (get-machine-version)))
-  *machine-version*)
+  (if (boundp '*machine-version*)
+      *machine-version*
+      (setf *machine-version*
+            (awhen (get-machine-version) (possibly-base-stringize it)))))
 
 ;;; FIXME: Don't forget to set these in a sample site-init file.
 ;;; FIXME: Perhaps the functions could be SETFable instead of having the
@@ -49,7 +43,7 @@ are running on, or NIL if we can't find any useful information."
 ;;; from ANSI 11.1.2.1.1 "Constraints on the COMMON-LISP Package
 ;;; for Conforming Implementations" it is kosher to add a SETF function for
 ;;; a symbol in COMMON-LISP..
-(declaim (type (or null string) *short-site-name* *long-site-name*))
+(declaim (type (or null simple-string) *short-site-name* *long-site-name*))
 (define-load-time-global *short-site-name* nil
   "The value of SHORT-SITE-NAME.")
 (define-load-time-global *long-site-name* nil
@@ -262,3 +256,24 @@ version 1[.0.0...] or greater."
   (declare (type (or null string) string))
   (push (list string name doc-type) sb-pcl::*!docstrings*)
   string)
+
+(in-package "SB-LOCKLESS")
+(defstruct (list-node
+            (:conc-name nil)
+            (:constructor %make-sentinel-node ())
+            (:copier nil))
+  (%node-next nil))
+
+;;; Specialized list variants will be created for
+;;;  fixnum, integer, real, string, generic "comparable"
+;;; but the node type and list type is the same regardless of key type.
+(defstruct (linked-list
+            (:constructor %make-lfl
+                          (head inserter deleter finder inequality equality))
+            (:conc-name list-))
+  (head       nil :type list-node :read-only t)
+  (inserter   nil :type function :read-only t)
+  (deleter    nil :type function :read-only t)
+  (finder     nil :type function :read-only t)
+  (inequality nil :type function :read-only t)
+  (equality   nil :type function :read-only t))
