@@ -44,6 +44,7 @@
               value2)))))
 
 (defvar *make-host-parallelism* nil)
+(defvar *fail-on-warnings* t)
 (defun make-host-1-parallelism () (car *make-host-parallelism*))
 (defun make-host-2-parallelism () (cdr *make-host-parallelism*))
 
@@ -163,24 +164,42 @@
     (rename-file x path)))
 (compile 'rename-file-a-la-unix)
 
-(export '(prepend-genfile-path read-from-file *generated-sources-root*))
+(export '(*target-sbcl-version* *generated-sources-root*
+          stem-source-path find-bootstrap-file read-from-file))
+(defvar *sources-root* "")
 (defvar *generated-sources-root* "")
+(defvar *src-cold-shared-pathname* *load-pathname*)
 
 ;;; See remark in COMPILE-STEM about strings vs. The Common Lisp Way
-(defun prepend-genfile-path (namestring)
-  (concatenate 'string
-               ;; if exact match to "output/", or mismatch at the next character
-               (if (member (mismatch "output/" namestring) '(nil 7))
-                   *generated-sources-root*
-                   "")
-               namestring))
-(compile 'prepend-genfile-path) ; seems in vogue to compile everything in this file
+(defun find-bootstrap-file (namestring)
+  (cond ((char= (char namestring 0) #\^)
+         ;; If it starts with a "^" then it means "src/cold/..."
+         (let ((this *src-cold-shared-pathname*)
+               (name (subseq namestring 1)))
+           (make-pathname :host (pathname-host this)
+                          :device (pathname-device this)
+                          :directory (pathname-directory this)
+                          :name (pathname-name name)
+                          :type (or (pathname-type name) (pathname-type this)))))
+        ((find #\/ namestring)
+         ;; Otherwise if it contains a slash, then it's a source file which is either
+         ;; in the tree as checked in, or generated by a prior build step.
+         (concatenate 'string
+                      (if (eql (mismatch "output/" namestring) 7) ; a generated source
+                          *generated-sources-root*
+                          *sources-root*)
+                      namestring))
+        (t
+         ;; Else, it's an optional user-supplied customization file,
+         ;; or a generated data file in the root directory such as "version.lisp-expr"
+         namestring)))
+(compile 'find-bootstrap-file) ; seems in vogue to compile everything in this file
 
 ;;; Return an expression read from the file named NAMESTRING.
 ;;; For user-supplied inputs, protect against more than one expression
-;;; appearing in the file. With trusted inputs we needn't bother.
+;;; appearing in the file. For in-tree inputs we needn't bother.
 (defun read-from-file (namestring &optional (enforce-single-expr t))
-  (with-open-file (s (prepend-genfile-path namestring))
+  (with-open-file (s (find-bootstrap-file namestring))
     (let* ((result (read s))
            (eof-result (cons nil nil)))
       (unless enforce-single-expr
@@ -201,7 +220,7 @@
 #+sbcl
 (progn
   (setq cl:*compile-print* nil)
-  (load "src/cold/muffler.lisp")
+  (load (find-bootstrap-file "^muffler"))
   ;; Let's just say we never care to see these.
   (declaim (sb-ext:muffle-conditions
             (satisfies unable-to-optimize-note-p)
@@ -211,7 +230,17 @@
 ;;;; special read-macros for building the cold system (and even for
 ;;;; building some of our tools for building the cold system)
 
-(load "src/cold/shebang.lisp")
+(load (find-bootstrap-file "^shebang"))
+
+;;; Subfeatures could be assigned as late as the beginning of make-host-2,
+;;; but I don't want to introduce another mechanism for delaying reading
+;;; of the customizer just because we can.
+;;; But it's not well-advertised; does it really merit a customization file?
+(export 'backend-subfeatures)
+(defvar backend-subfeatures
+  (let ((customizer-file-name "customize-backend-subfeatures.lisp"))
+    (when (probe-file customizer-file-name)
+      (copy-list (funcall (compile nil (read-from-file customizer-file-name)) nil)))))
 
 ;;; When cross-compiling, the *FEATURES* set for the target Lisp is
 ;;; not in general the same as the *FEATURES* set for the host Lisp.
@@ -227,57 +256,54 @@
 ;;; The compromise is to examine a variable specifying a path
 ;;; (and it can't go in SB-COLD because the package is not made soon enough)
 (setf sb-xc:*features*
-      (let* ((pathname (let ((var 'cl-user::*sbcl-target-features-file*))
+      (let* ((pathname (let ((var 'cl-user::*sbcl-local-target-features-file*))
                          (if (boundp var)
                              (symbol-value var)
                              "local-target-features.lisp-expr")))
              (default-features
                (funcall (compile nil (read-from-file pathname))
-                        (read-from-file "base-target-features.lisp-expr")))
+                        (read-from-file "^base-target-features.lisp-expr")))
              (customizer-file-name "customize-target-features.lisp")
              (customizer (if (probe-file customizer-file-name)
                              (compile nil
                                       (read-from-file customizer-file-name))
                              #'identity))
-             (target-feature-list (funcall customizer default-features))
+             ;; Bind temporarily so that TARGET-FEATUREP and TARGET-PLATFORM-KEYWORD
+             ;; can see the tentative list.
+             (sb-xc:*features* (funcall customizer default-features))
              (gc (find-if (lambda (x) (member x '(:cheneygc :gencgc)))
-                          target-feature-list))
-             (arch (target-platform-keyword target-feature-list)))
-        (when (featurep '(:and :sb-thread (:or :linux :win32)) target-feature-list)
-          (pushnew :sb-futex target-feature-list))
-        (when (featurep '(:and :sb-thread (:not :win32)) target-feature-list)
-          (push :pauseless-threadstart target-feature-list))
-        (when (featurep '(:and :sb-thread (:or :darwin :openbsd)) target-feature-list)
-          (push :os-thread-stack target-feature-list))
-        (when (and (member :x86 target-feature-list)
-                   (member :int4-breakpoints target-feature-list))
+                          sb-xc:*features*))
+             (arch (target-platform-keyword)))
+        ;; Win32 conditionally adds :sb-futex in grovel-features.sh
+        (when (target-featurep '(:and :sb-thread (:or :linux :freebsd)))
+          (pushnew :sb-futex sb-xc:*features*))
+        (when (target-featurep '(:and :sb-thread (:not :win32)))
+          (push :pauseless-threadstart sb-xc:*features*))
+        (when (target-featurep '(:and :sb-thread (:or :darwin :openbsd)))
+          (push :os-thread-stack sb-xc:*features*))
+        (when (target-featurep '(:and :x86 :int4-breakpoints))
           ;; 0xCE is a perfectly good 32-bit instruction,
           ;; unlike on x86-64 where it is illegal. It's therefore
           ;; confusing to allow this feature in a 32-bit build.
           ;; But it's annoying to have a build script that otherwise works
           ;; for a native x86/x86-64 build except for needing one change.
           ;; Just print something and go on with life.
-          (setq target-feature-list
-                (remove :int4-breakpoints target-feature-list))
+          (setq sb-xc:*features* (remove :int4-breakpoints sb-xc:*features*))
           (warn "Removed :INT4-BREAKPOINTS from target features"))
+        (when (target-featurep '(:or :arm64 :sse4))
+          (push :round-float sb-xc:*features*))
+        (when (target-featurep '(:and :arm64 :darwin))
+          (push :arm-v8.1 backend-subfeatures))
+
         ;; Putting arch and gc choice first is visually convenient, versus
         ;; having to parse a random place in the line to figure out the value
         ;; of a binary choice {cheney vs gencgc} and architecture.
         ;; De-duplicate the rest of the symbols because the command line
         ;; can add redundant --with-mumble options.
         (list* arch gc (sort (remove-duplicates
-                              (remove arch (remove gc target-feature-list)))
+                              (remove arch (remove gc sb-xc:*features*)))
                              #'string<))))
 
-(defvar *shebang-backend-subfeatures*
-  (let* ((default-subfeatures nil)
-         (customizer-file-name "customize-backend-subfeatures.lisp")
-         (customizer (if (probe-file customizer-file-name)
-                         (compile nil
-                                  (read-from-file customizer-file-name))
-                         #'identity)))
-    (funcall customizer default-subfeatures)))
-
 ;;; Call for effect of signaling an error if no target picked.
 (target-platform-keyword)
 
@@ -295,20 +321,19 @@
           ":SB-THREAD requires :GENCGC")
          ("(and pauseless-threadstart (not sb-thread))"
           ":PAUSELESS-THREADSTART requires :SB-THREAD")
+         ("(and sb-safepoint (not sb-thread))" ":SB-SAFEPOINT requires :SB-THREAD")
          ("(and sb-thread (not (or riscv ppc ppc64 x86 x86-64 arm64)))"
           ":SB-THREAD not supported on selected architecture")
          ("(and gencgc cheneygc)"
           ":GENCGC and :CHENEYGC are incompatible")
-         ("(and cheneygc (not (or alpha arm arm64 hppa mips ppc riscv sparc)))"
+         ("(and cheneygc (not (or arm mips ppc riscv sparc)))"
           ":CHENEYGC not supported on selected architecture")
          ("(and gencgc (not (or sparc ppc ppc64 x86 x86-64 arm arm64 riscv)))"
           ":GENCGC not supported on selected architecture")
          ("(not (or gencgc cheneygc))"
           "One of :GENCGC or :CHENEYGC must be enabled")
-         ("(and sb-safepoint (not (or arm64 ppc x86 x86-64)))"
-          ":SB-SAFEPOINT not supported on selected architecture")
-         ("(and sb-safepoint-strictly (not sb-safepoint))"
-          ":SB-SAFEPOINT-STRICTLY requires :SB-SAFEPOINT")
+         ("(and sb-safepoint (not (and (or arm64 x86 x86-64) (or darwin linux win32))))"
+          ":SB-SAFEPOINT not supported on selected arch/OS")
          ("(not (or elf mach-o win32))"
           "No execute object file format feature defined")
          ("(and cons-profiling (not sb-thread))" ":CONS-PROFILING requires :SB-THREAD")
@@ -340,7 +365,7 @@
         (push (second test) failed-test-descriptions))))
   (when failed-test-descriptions
     (error "Feature compatibility check failed, ~S"
-           failed-test-descriptions)))
+           (reverse failed-test-descriptions))))
 
 ;;;; cold-init-related PACKAGE and SYMBOL tools
 
@@ -352,7 +377,7 @@
 ;;; All code depending on this is itself dependent on #+SB-SHOW.
 (defvar *cl-snapshot*)
 (when (member :sb-show sb-xc:*features*)
-  (load "src/cold/snapshot.lisp")
+  (load (find-bootstrap-file "^snapshot"))
   (setq *cl-snapshot* (take-snapshot "COMMON-LISP")))
 
 ;;;; master list of source files and their properties
@@ -419,8 +444,8 @@
              (,flags (rest ,stem-and-flags)))
          ,@body))))
 
-;;; Given a STEM, remap the path component "/target/" to a suitable
-;;; target directory.
+;;; Given a STEM, remap the path components "/{arch}/" and "/asm-target/"
+;;; to suitable directories.
 (defun stem-remap-target (stem)
   (flet ((try-replacing (this that)
            (let ((position (search this stem)))
@@ -429,20 +454,20 @@
                             (subseq stem 0 (1+ position))
                             (string-downcase that)
                             (subseq stem (+ position (length this) -1)))))))
-    (or (try-replacing "/target/" (target-platform-keyword))
+    (or (try-replacing "/{arch}/" (target-platform-keyword))
         (try-replacing "/asm-target/" (backend-assembler-target-name))
         stem)))
 (compile 'stem-remap-target)
 
 ;;; Determine the source path for a stem by remapping from the abstract name
-;;; if it contains "/target/" and appending a ".lisp" suffix.
+;;; if it contains "/{arch}/" and appending a ".lisp" suffix.
 ;;; Assume that STEM is source-tree-relative unless it starts with "output/"
 ;;; in which case it could be elsewhere, if you prefer to keep the sources
 ;;; devoid of compilation artifacts. (The production of out-of-tree artifacts
 ;;; is not actually implemented in the generic build, however if your build
 ;;; system does that by itself, then hooray for you)
 (defun stem-source-path (stem)
-  (concatenate 'string (prepend-genfile-path (stem-remap-target stem)) ".lisp"))
+  (concatenate 'string (find-bootstrap-file (stem-remap-target stem)) ".lisp"))
 (compile 'stem-source-path)
 
 ;;; Determine the object path for a stem/flags/mode combination.
@@ -482,7 +507,7 @@
           ;; to produce warnings as a bug workaround.
           (let ((cl:*features* (cons feature cl:*features*))
                 (*readtable* *xc-readtable*))
-            (read-from-file "build-order.lisp-expr" nil))))
+            (read-from-file "^build-order.lisp-expr" nil))))
     (setf *stems-and-flags* (cons build-phase list)))
   ;; Now check for duplicate stems and bogus flags.
   (let ((stems (make-hash-table :test 'equal)))
@@ -539,7 +564,18 @@
                        (if *compile-for-effect-only* "-scratch" "-tmp")))
 
          (compile-file (ecase mode
-                         (:host-compile #'compile-file)
+                         (:host-compile
+                          #+ccl
+                          (lambda (&rest args)
+                            ;; CCL doesn't like NOTINLINE on uknown functions
+                            (handler-bind ((ccl:compiler-warning
+                                             (lambda (c)
+                                               (when (eq (ccl::compiler-warning-warning-type c)
+                                                         :unknown-declaration-function)
+                                                 (muffle-warning c)))))
+                              (apply #'compile-file args)))
+                          #-ccl
+                          #'compile-file)
                          (:target-compile (if (find :assem flags)
                                               *target-assemble-file*
                                               *target-compile-file*))))
@@ -724,6 +760,8 @@
 ;;;; Floating-point number reader interceptor
 
 (defvar *choke-on-host-irrationals* t)
+;;; FIXME: this gets stuck on forms which contain literal CTYPE objects
+;;; because of infinite recursion.
 (defun install-read-interceptor ()
   ;; Intercept READ to catch inadvertent use of host floating-point literals.
   ;; This prevents regressions in the portable float logic and allows passing
@@ -761,7 +799,7 @@
                ((and structure-object (not package))
                 (let ((type-name (string (type-of x))))
                   ;; This "LAYOUT" refers to *our* object, not host-sb-kernel:layout.
-                  (unless (member type-name '("LAYOUT" "FLOAT" "COMPLEXNUM")
+                  (unless (member type-name '("WRAPPER" "LAYOUT" "FLOAT" "COMPLEXNUM")
                                   :test #'string=)
                     ;(Format t "visit a ~/host-sb-ext:print-symbol-with-prefix/~%" (type-of x))
                     ;; This generalizes over any structure. I need it because we
