@@ -14,6 +14,10 @@
 
 (declaim (special *compiler-error-bailout*))
 
+;;; the lexical environment we are currently converting in
+(defvar *lexenv*)
+(declaim (type lexenv *lexenv*))
+
 ;;; *CURRENT-FORM-NUMBER* is used in FIND-SOURCE-PATHS to compute the
 ;;; form number to associate with a source path. This should be bound
 ;;; to an initial value of 0 before the processing of each truly
@@ -64,6 +68,11 @@
     (setf (gethash form *source-paths*)
           (apply #'list* 'original-source-start *current-form-number* arguments))))
 
+(defun proper-list (form)
+  (if (proper-list-p form)
+      form
+      (compiler-error "~@<~S is not a proper list.~@:>" form)))
+
 ;;; *CURRENT-COMPONENT* is the COMPONENT structure which we link
 ;;; blocks into as we generate them. This just serves to glue the
 ;;; emitted blocks together until local call analysis and flow graph
@@ -100,6 +109,34 @@
 
 ;;;; namespace management utilities
 
+;;; Unpack (INFO :FUNCTION :INLINING-DATA FUN-NAME). If an explicit expansion
+;;; is not stored but FUN-NAME is a structure constructor, reconstitute the
+;;; expansion from the defstruct description. Secondary value is T if the expansion
+;;; was explicit. See SAVE-INLINE-EXPANSION-P for why we care.
+;;; (Essentially, once stored then always stored, lest inconsistency result)
+;;; If we have just a DXABLE-ARGS, or nothing at all, return NIL and NIL.
+;;; If called on a string or anything that is not a function designator,
+;;; return NIL and NIL.
+(declaim (ftype (sfunction (t) (values list boolean))
+                fun-name-inline-expansion))
+(defun fun-name-inline-expansion (fun-name)
+  (multiple-value-bind (answer winp) (info :function :inlining-data fun-name)
+    (typecase answer
+      ;; an INLINING-DATA is a DXABLE-ARGS, so test it first
+      (inlining-data (setq answer (inlining-data-expansion answer)))
+      (dxable-args   (setq answer nil winp nil)))
+    (when (and (not winp) (symbolp fun-name))
+      (let ((info (info :function :source-transform fun-name)))
+        (when (typep info '(cons defstruct-description (eql :constructor)))
+          (let* ((dd (car info)) (spec (assq fun-name (dd-constructors dd))))
+            (aver spec)
+            (setq answer `(lambda ,@(structure-ctor-lambda-parts dd (cdr spec))))))))
+    (values answer winp)))
+(defun fun-name-dx-args (fun-name)
+  (let ((answer (info :function :inlining-data fun-name)))
+    (when (typep answer 'dxable-args)
+      (dxable-args-list answer))))
+
 ;; As with LEXENV-FIND, we assume use of *LEXENV*, but macroexpanders
 ;; receive an explicit environment and should pass it.
 ;; A declaration will trump a proclamation.
@@ -121,7 +158,7 @@
 
 #-sb-devel
 (declaim (start-block find-free-fun find-lexically-apparent-fun
-                      ;; needed by with-fun-name-leaf
+                      ;; needed by ir1-translators
                       find-global-fun))
 
 (defun maybe-defined-here (name where)
@@ -228,7 +265,17 @@
   (or (let ((old-free-fun (gethash name free-funs)))
         (when old-free-fun
           (clear-invalid-functionals old-free-fun)
-          old-free-fun))
+          (when (or (not (defined-fun-p old-free-fun))
+                    (not (block-compile *compilation*))
+                    ;; When block-compiling, it is the case that we
+                    ;; could proclaim an inline, define some things,
+                    ;; then proclaim a notinline, and in the case that
+                    ;; the function's :inlinep info changes, the
+                    ;; old-free-fun is stale and we should make a new
+                    ;; one.
+                    (eq (info :function :inlinep name)
+                        (defined-fun-inlinep old-free-fun)))
+            old-free-fun)))
       (let ((kind (info :function :kind name)))
         (ecase kind
           ((:macro :special-form)
@@ -324,8 +371,8 @@
 ;;; processed with MAKE-LOAD-FORM. We have to be careful, because
 ;;; CONSTANT might be circular. We also check that the constant (and
 ;;; any subparts) are dumpable at all.
-(defun ensure-externalizable (constant)
-  (declare (inline alloc-xset))
+(defun maybe-emit-make-load-forms (constant)
+  (declare #-sb-xc-host (inline alloc-xset))
   (dx-let ((xset (alloc-xset)))
     (named-let grovel ((value constant))
       ;; Unless VALUE is an object which which can't contain other objects,
@@ -346,30 +393,30 @@
            (dotimes (i (length value))
              (grovel (aref value i))))
           ((simple-array t)
-                    ;; Even though the (ARRAY T) branch does the exact
-                    ;; same thing as this branch we do this separately
-                    ;; so that the compiler can use faster versions of
-                    ;; array-total-size and row-major-aref.
+           ;; Even though the (ARRAY T) branch does the exact
+           ;; same thing as this branch we do this separately
+           ;; so that the compiler can use faster versions of
+           ;; array-total-size and row-major-aref.
            (dotimes (i (array-total-size value))
              (grovel (row-major-aref value i))))
           ((array t)
            (dotimes (i (array-total-size value))
              (grovel (row-major-aref value i))))
           (instance
-                    ;; In the target SBCL, we can dump any instance, but
-                    ;; in the cross-compilation host, %INSTANCE-FOO
-                    ;; functions don't work on general instances, only on
-                    ;; STRUCTURE!OBJECTs.
-                    ;;
-                    ;; Behold the wonderfully clear sense of this-
-                    ;;  WHEN (EMIT-MAKE-LOAD-FORM VALUE)
-                    ;; meaning "when you're _NOT_ using a custom load-form"
-                    ;;
-                    ;; FIXME: What about funcallable instances with
-                    ;; user-defined MAKE-LOAD-FORM methods?
+           ;; In the target SBCL, we can dump any instance, but
+           ;; in the cross-compilation host, %INSTANCE-FOO
+           ;; functions don't work on general instances, only on
+           ;; STRUCTURE!OBJECTs.
+           ;;
+           ;; Behold the wonderfully clear sense of this-
+           ;;  WHEN (EMIT-MAKE-LOAD-FORM VALUE)
+           ;; meaning "when you're _NOT_ using a custom load-form"
+           ;;
+           ;; FIXME: What about funcallable instances with
+           ;; user-defined MAKE-LOAD-FORM methods?
            (when (emit-make-load-form value)
              #+sb-xc-host
-             (aver (eql (layout-bitmap (%instance-layout value))
+             (aver (eql (wrapper-bitmap (%instance-wrapper value))
                         sb-kernel:+layout-all-tagged+))
              (do-instance-tagged-slot (i value)
                (grovel (%instance-ref value i)))))
@@ -382,7 +429,7 @@
 ;;; A constant is trivially externalizable if it involves no INSTANCE types
 ;;; or any un-dumpable object.
 (defun trivially-externalizable-p (constant)
-  (declare (inline alloc-xset))
+  (declare #-sb-xc-host (inline alloc-xset))
   (dx-let ((xset (alloc-xset)))
     (named-let ok ((value constant))
       (if (or (dumpable-leaflike-p value) (xset-member-p value xset))
@@ -439,16 +486,6 @@
 
 ;;; Insert NEW before OLD in the flow-graph.
 (defun insert-node-before (old new)
-  (let ((prev (node-prev old))
-        (temp (make-ctran)))
-    (ensure-block-start prev)
-    (setf (ctran-next prev) nil)
-    (link-node-to-previous-ctran new prev)
-    (use-ctran new temp)
-    (link-node-to-previous-ctran old temp))
-  (values))
-
-(defun insert-node-before-no-split (old new)
   (let ((prev (node-prev old))
         (temp (make-ctran)))
     (setf (ctran-next prev) nil)
@@ -611,8 +648,6 @@
                       (reference-leaf start next result form))
                      (t
                       (reference-constant start next result form))))
-              ((not (proper-list-p form))
-               (compiler-error "~@<~S is not a proper list.~@:>" form))
               (t
                (ir1-convert-functoid start next result form)))))
     (values))
@@ -801,7 +836,7 @@
   (let* ((op (car form))
          (translator (and (symbolp op) (info :function :ir1-convert op))))
     (if translator
-        (funcall translator start next result form)
+        (funcall translator start next result (proper-list form))
         (multiple-value-bind (res cmacro-fun-name)
                (expand-compiler-macro form)
              (cond ((eq res form)
@@ -959,30 +994,13 @@
             ;; If this source path has already been instrumented in
             ;; this block, don't instrument it again.
             start
-            (let ((store
-                   ;; Get an interned record cons for the path. A cons
-                   ;; with the same object identity must be used for
-                   ;; each instrument for the same block.
-                   (ensure-gethash path (code-coverage-records metadata)
-                                   (cons path +code-coverage-unmarked+)))
-                  (next (make-ctran))
+            (let ((next (make-ctran))
                   (*allow-instrumenting* nil))
-              #+(or x86-64 x86) (declare (ignore store)) ; eval'd for side-effect only
+              (ensure-gethash path (code-coverage-records metadata)
+                              (cons path +code-coverage-unmarked+))
               (push (ctran-block start)
                     (gethash path (code-coverage-blocks metadata)))
-              (ir1-convert start next nil
-                           #+(or x86-64 x86)
-                           `(%primitive mark-covered ',path)
-                           #-(or x86-64 x86)
-                           `(locally
-                                (declare (optimize speed
-                                                   (safety 0)
-                                                   (debug 0)
-                                                   (check-constant-modification 0)))
-                              ;; We're being naughty here, and
-                              ;; modifying constant data. That's ok,
-                              ;; we know what we're doing.
-                              (%rplacd ',store t)))
+              (ir1-convert start next nil `(%primitive mark-covered ',path))
               next)))
       start))
 
@@ -1037,7 +1055,7 @@
     (ir1-convert start ctran fun-lvar `(the (or function symbol) ,fun))
     (let ((combination
            (ir1-convert-combination-args fun-lvar ctran next result
-                                         (cdr form))))
+                                         (cdr (proper-list form)))))
       (when (step-form-p form)
         ;; Store a string representation of the form in the
         ;; combination node. This will let the IR2 translator know
@@ -1083,8 +1101,13 @@
         (setf (combination-args node) (arg-lvars))))
     node))
 
+(defun show-transform-p (showp fun-name)
+  (or (and (listp showp) (member fun-name showp :test 'equal))
+      (eq showp t)))
 (defun show-transform (kind name new-form)
-  (let ((*print-right-margin* 100))
+  (let ((*print-length* 100)
+        (*print-level* 50)
+        (*print-right-margin* 128))
     (format *trace-output* "~&xform (~a) ~S~% -> ~S~%"
             kind name new-form)))
 ;;; Convert a call to a global function. If not NOTINLINE, then we do
@@ -1108,15 +1131,20 @@
                       (struct-fun-transform transform form name))
                 ;; Note that "pass" means fail. Gotta love it.
                 (cond (pass
-                       (ir1-convert-maybe-predicate start next result form var))
+                       (ir1-convert-maybe-predicate start next result
+                                                    (proper-list form)
+                                                    var))
                       (t
                        (unless (policy *lexenv* (zerop store-xref-data))
                          (record-call name (ctran-block start) *current-path*))
-                       (when *show-transforms-p*
+                       (when (show-transform-p *show-transforms-p* name)
                          (show-transform "src" name transformed))
                        (let ((*transforming* t))
                          (ir1-convert start next result transformed)))))
-              (ir1-convert-maybe-predicate start next result form var))))))
+              (ir1-convert-maybe-predicate start next result
+                                           (proper-list form)
+                                           var))))))
+
 
 ;;; KLUDGE: If we insert a synthetic IF for a function with the PREDICATE
 ;;; attribute, don't generate any branch coverage instrumentation for it.
