@@ -9,16 +9,79 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;; a value for an optimization declaration
-(def!type policy-quality () '(integer 0 3))
+(deftype policy-quality () '(integer 0 3))
 
+(defconstant-eqx +policy-primary-qualities+
+        #(;; ANSI standard qualities
+          compilation-speed
+          debug
+          safety
+          space
+          speed
+          ;; SBCL extensions
+          ;;
+          ;; FIXME: INHIBIT-WARNINGS is a misleading name for this.
+          ;; Perhaps BREVITY would be better. But the ideal name would
+          ;; have connotations of suppressing not warnings but only
+          ;; optimization-related notes, which is already mostly the
+          ;; behavior, and should probably become the exact behavior.
+          ;; Perhaps INHIBIT-NOTES?
+          inhibit-warnings)
+    #'equalp)
+
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (defconstant n-policy-primary-qualities (length +policy-primary-qualities+))
+  ;; 1 bit per quality is stored to indicate whether it was explicitly given
+  ;; a value in a lexical policy. In addition to the 5 ANSI-standard qualities,
+  ;; SBCL defines one more "primary" quality and 16 dependent qualities.
+  ;; Both kinds take up 1 bit in the mask of specified qualities.
+  (defconstant max-policy-qualities 32))
+
+;; Each primary and dependent quality policy is assigned a small integer index.
+;; The POLICY struct represents a set of policies in an order-insensitive way
+;; that facilitates quicker lookup than scanning an alist.
+(defstruct (policy (:constructor make-policy
+                       (primary-qualities &optional
+                                          presence-bits dependent-qualities)))
+  ;; Mask with a 1 for each quality that has an explicit value in this policy.
+  ;; Primary qualities fill the mask from left-to-right and dependent qualities
+  ;; from right-to-left.
+  ;; xc has trouble folding this MASK-FIELD, but it works when host-evaluated.
+  (presence-bits #.(mask-field
+                    (byte n-policy-primary-qualities
+                          (- max-policy-qualities n-policy-primary-qualities))
+                    -1)
+                 :type (unsigned-byte #.max-policy-qualities))
+  ;; For efficiency, primary qualities are segregated because there are few
+  ;; enough of them to fit in a fixnum.
+  (primary-qualities 0 :type (unsigned-byte #.(* 2 n-policy-primary-qualities)))
+  ;; 2 bits per dependent quality is a fixnum on 64-bit build, not on 32-bit.
+  ;; It would certainly be possible to constrain this to storing exactly
+  ;; the 16 currently defined dependent qualities,
+  ;; but that would be overly limiting.
+  (dependent-qualities 0
+   :type (unsigned-byte #.(* (- max-policy-qualities n-policy-primary-qualities)
+                             2))))
+(declaim (freeze-type policy))
+
+;;; *POLICY* holds the current global compiler policy information, as
+;;; a POLICY object mapping from the compiler-assigned index (unique per
+;;; quality name) to quality value.
+;;; This used to be an alist, but tail-sharing was never really possible
+;;; because for deterministic comparison the list was always freshly
+;;; consed so that destructive sorting could be done for canonicalization.
+(defvar *policy*)
 (defvar *macro-policy* nil)
 ;;; global policy restrictions as a POLICY object or nil
 (defvar *policy-min* nil)
 (defvar *policy-max* nil)
 
+(declaim (type policy *policy*)
+         (type (or policy null) *policy-min* *policy-max*))
+
 (defun restrict-compiler-policy (&optional quality (min 0) (max 3))
   "Assign a minimum value to an optimization quality. QUALITY is the name of
 the optimization quality to restrict, MIN (defaulting to zero) is the
@@ -75,6 +138,7 @@ See also :POLICY option in WITH-COMPILATION-UNIT."
   (getter nil :read-only t)
   (documentation nil :read-only t)
   (values-documentation nil :read-only t))
+(declaim (freeze-type policy-dependent-quality))
 
 ;;; names of recognized optimization policy qualities
 (declaim (simple-vector **policy-dependent-qualities**))
@@ -147,15 +211,6 @@ See also :POLICY option in WITH-COMPILATION-UNIT."
     (otherwise
      nil)))
 
-;;; *POLICY* holds the current global compiler policy information, as
-;;; a POLICY object mapping from the compiler-assigned index (unique per
-;;; quality name) to quality value.
-;;; This used to be an alist, but tail-sharing was never really possible
-;;; because for deterministic comparison the list was always freshly
-;;; consed so that destructive sorting could be done for canonicalization.
-(declaim (type policy *policy*)
-         (type (or policy null) *policy-min* *policy-max*))
-
 ;; ANSI-specified default of 1 for each quality.
 (defglobal **baseline-policy** nil)
 ;; Baseline policy altered with (TYPE-CHECK 0)
@@ -187,11 +242,11 @@ See also :POLICY option in WITH-COMPILATION-UNIT."
                            sum (ash #b01 (* i 2))))
         *policy* (copy-policy **baseline-policy**))
   (when baseline-qualities
-    (sb!xc:proclaim `(optimize ,@baseline-qualities))
+    (proclaim `(optimize ,@baseline-qualities))
     ;; Copy altered policy back as the baseline policy
     (setq **baseline-policy** (copy-policy *policy*)))
   (let ((*policy* *policy*))
-    (sb!xc:proclaim '(optimize (type-check 0)))
+    (proclaim '(optimize (type-check 0)))
     (setq **zero-typecheck-policy** *policy*)))
 
 ;;; Look up a named optimization quality in POLICY. This is only
@@ -278,18 +333,17 @@ See also :POLICY option in WITH-COMPILATION-UNIT."
 ;;; THING. EXPR is a form which accesses optimization qualities by
 ;;; referring to them by name, e.g. (> SPEED SPACE).
 (defmacro policy (thing expr &optional (coercion-fn '%coerce-to-policy))
-  (declare (notinline identity)) ; constant-folding-error otherwise. (FIXME)
   (let* ((n-policy (make-symbol "P"))
-         (binds (loop for name across (identity +policy-primary-qualities+)
-                      for index downfrom -1
+         (binds (loop for index downfrom -1
+                      for name across +policy-primary-qualities+
                       collect `(,name (%policy-quality ,n-policy ,index))))
          (dependent-binds
-          (loop for info across **policy-dependent-qualities**
-                for name = (policy-dependent-quality-name info)
-               collect `(,name (let ((,name (policy-quality ,n-policy ',name)))
-                                 (if (= ,name 1)
-                                     ,(policy-dependent-quality-expression info)
-                                     ,name))))))
+           (loop for info across **policy-dependent-qualities**
+                 for name = (policy-dependent-quality-name info)
+                 collect `(,name (let ((,name (policy-quality ,n-policy ',name)))
+                                   (if (= ,name 1)
+                                       ,(policy-dependent-quality-expression info)
+                                       ,name))))))
     `(let ((,n-policy (,coercion-fn ,thing)))
        ;; FIXME: automatically inserted IGNORABLE decls are
        ;; often suggestive of poor style, as is this one.
@@ -395,16 +449,10 @@ See also :POLICY option in WITH-COMPILATION-UNIT."
 ;; Same for {UN}MUFFLE-CONDITIONS or anything else that can be proclaimed.
 ;;
 (defun set-macro-policy (list)
-  ;: Note that *MACRO-POLICY* does not represent absence of any primary quality,
-  ;; and therefore whenever it is injected into a macro, you get all baseline
-  ;; values of 1, augmented by the specified changes.
-  ;; There are two alternative behaviors that might make sense:
-  ;; - use the value of *POLICY* when SET-MACRO-POLICY is called as the baseline
-  ;;   augmented by the specifiers in LIST
-  ;; - use the lexical policy at the time of expansion, augmented by LIST
-  ;; But most probably the current behavior is entirely reasonable.
+  ;; Unlike most constructs that involve a policy, assume total absence of all
+  ;; standard qualities. Record only the qualities expressed in LIST.
   (setq *macro-policy* (process-optimize-decl `(optimize ,@list)
-                                              **baseline-policy**)))
+                                              (make-policy 0 0 0))))
 
 ;; Turn the macro policy into an OPTIMIZE declaration for insertion
 ;; into a macro body for DEFMACRO, MACROLET, or DEFINE-COMPILER-MACRO.
