@@ -229,7 +229,7 @@
   (:variant funcallable-instance-info-offset fun-pointer-lowtag)
   (:translate %funcallable-instance-info))
 
-(define-vop (set-funcallable-instance-info word-index-set)
+(define-vop (set-funcallable-instance-info word-index-set-nr)
   (:variant funcallable-instance-info-offset fun-pointer-lowtag)
   (:translate %set-funcallable-instance-info))
 
@@ -279,7 +279,7 @@
   (:variant instance-slots-offset instance-pointer-lowtag)
   (:arg-types * positive-fixnum))
 
-(define-vop (instance-index-set word-index-set)
+(define-vop (instance-index-set word-index-set-nr)
   (:policy :fast-safe)
   (:translate %instance-set)
   (:variant instance-slots-offset instance-pointer-lowtag)
@@ -324,18 +324,15 @@
                         (index :scs (any-reg))
                         (value :scs (,sc)))
                  (:arg-types * positive-fixnum ,result-type)
-                 (:results (result :scs (,sc)))
                  (:temporary (:scs (non-descriptor-reg)) offset)
-                 (:result-types ,result-type)
                  (:generator 5
                    (inst add offset index (- (ash instance-slots-offset word-shift)
                                              instance-pointer-lowtag))
-                   (inst st value object offset)
-                   (move result value))))))
+                   (inst st value object offset))))))
   (def "" unsigned-reg unsigned-num)
   (def "SIGNED-" signed-reg signed-num))
 
-(define-vop (raw-instance-ref/single)
+(define-vop ()
   (:translate %raw-instance-ref/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -349,24 +346,20 @@
                               instance-pointer-lowtag))
     (inst ldf value object offset)))
 
-(define-vop (raw-instance-set/single)
+(define-vop ()
   (:translate %raw-instance-set/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (single-reg) :target result))
+         (value :scs (single-reg)))
   (:arg-types * positive-fixnum single-float)
-  (:results (result :scs (single-reg)))
-  (:result-types single-float)
   (:temporary (:scs (non-descriptor-reg)) offset)
   (:generator 5
     (inst add offset index (- (ash instance-slots-offset word-shift)
                               instance-pointer-lowtag))
-    (inst stf value object offset)
-    (unless (location= result value)
-      (inst fmovs result value))))
+    (inst stf value object offset)))
 
-(define-vop (raw-instance-ref/double)
+(define-vop ()
   (:translate %raw-instance-ref/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -380,24 +373,20 @@
                               instance-pointer-lowtag))
     (inst lddf value object offset)))
 
-(define-vop (raw-instance-set/double)
+(define-vop ()
   (:translate %raw-instance-set/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (double-reg) :target result))
+         (value :scs (double-reg)))
   (:arg-types * positive-fixnum double-float)
-  (:results (result :scs (double-reg)))
-  (:result-types double-float)
   (:temporary (:scs (non-descriptor-reg)) offset)
   (:generator 5
     (inst add offset index (- (ash instance-slots-offset word-shift)
                               instance-pointer-lowtag))
-    (inst stdf value object offset)
-    (unless (location= result value)
-      (move-double-reg result value))))
+    (inst stdf value object offset)))
 
-(define-vop (raw-instance-ref/complex-single)
+(define-vop ()
   (:translate %raw-instance-ref/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -413,32 +402,22 @@
     (inst add offset offset n-word-bytes)
     (inst ldf (complex-single-reg-imag-tn value) object offset)))
 
-(define-vop (raw-instance-set/complex-single)
+(define-vop ()
   (:translate %raw-instance-set/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (complex-single-reg) :target result))
+         (value :scs (complex-single-reg)))
   (:arg-types * positive-fixnum complex-single-float)
-  (:results (result :scs (complex-single-reg)))
-  (:result-types complex-single-float)
   (:temporary (:scs (non-descriptor-reg)) offset)
   (:generator 5
     (inst add offset index (- (ash instance-slots-offset word-shift)
                               instance-pointer-lowtag))
-    (let ((value-real (complex-single-reg-real-tn value))
-          (result-real (complex-single-reg-real-tn result)))
-      (inst stf value-real object offset)
-      (unless (location= result-real value-real)
-        (inst fmovs result-real value-real)))
+    (inst stf (complex-single-reg-real-tn value) object offset)
     (inst add offset offset n-word-bytes)
-    (let ((value-imag (complex-single-reg-imag-tn value))
-          (result-imag (complex-single-reg-imag-tn result)))
-      (inst stf value-imag object offset)
-      (unless (location= result-imag value-imag)
-        (inst fmovs result-imag value-imag)))))
+    (inst stf (complex-single-reg-imag-tn value) object offset)))
 
-(define-vop (raw-instance-ref/complex-double)
+(define-vop ()
   (:translate %raw-instance-ref/complex-double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
@@ -454,27 +433,17 @@
     (inst add offset offset (* 2 n-word-bytes))
     (inst lddf (complex-double-reg-imag-tn value) object offset)))
 
-(define-vop (raw-instance-set/complex-double)
+(define-vop ()
   (:translate %raw-instance-set/complex-double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (complex-double-reg) :target result))
+         (value :scs (complex-double-reg)))
   (:arg-types * positive-fixnum complex-double-float)
-  (:results (result :scs (complex-double-reg)))
-  (:result-types complex-double-float)
   (:temporary (:scs (non-descriptor-reg)) offset)
   (:generator 5
     (inst add offset index (- (ash instance-slots-offset word-shift)
                               instance-pointer-lowtag))
-    (let ((value-real (complex-double-reg-real-tn value))
-          (result-real (complex-double-reg-real-tn result)))
-      (inst stdf value-real object offset)
-      (unless (location= result-real value-real)
-        (move-double-reg result-real value-real)))
+    (inst stdf (complex-double-reg-real-tn value) object offset)
     (inst add offset offset (* 2 n-word-bytes))
-    (let ((value-imag (complex-double-reg-imag-tn value))
-          (result-imag (complex-double-reg-imag-tn result)))
-      (inst stdf value-imag object offset)
-      (unless (location= result-imag value-imag)
-        (move-double-reg result-imag value-imag)))))
+    (inst stdf (complex-double-reg-imag-tn value) object offset)))
