@@ -46,8 +46,7 @@
           (cons (cons type name)
                 (remove name *backend-type-predicates*
                         :key #'cdr)))
-    (%deftransform name '(function (t) *) #'fold-type-predicate)
-    name))
+    (%deftransform name nil '(function (t) *) #'fold-type-predicate)))
 
 ;;;; IR1 transforms
 
@@ -358,14 +357,29 @@
 (defun source-transform-hairy-typep (object type)
   (declare (type hairy-type type))
   (let ((spec (hairy-type-specifier type)))
-    (cond ((unknown-type-p type)
+    (cond ((and (unknown-type-p type)
+                (symbolp spec)
+                (eq (info :type :kind spec) :forthcoming-defclass-type))
+           ;; Knowing that it was DEFCLASSed is enough to emit a CLASSOID-CELL-TYPEP test.
+           ;; Combinators involving this - e.g. (OR A-NEW-CLASS OTHER-CLASS) -
+           ;; are handled correctly, because we don't punt on everything in the expression
+           ;; as soon as any unknown is present.
+           `(classoid-cell-typep ,(find-classoid-cell spec :create t) ,object))
+          ((unknown-type-p type)
            #+sb-xc-host
            (warn "can't open-code test of unknown type ~S"
                  (type-specifier type))
-           #-sb-xc-host
-           (when (policy *lexenv* (> speed inhibit-warnings))
-             (compiler-notify "can't open-code test of unknown type ~S"
-                              (type-specifier type)))
+           ;; This is not a policy-based decision to notify here,
+           ;; because it is _ALWAYS_ questionable style imho to refer to unknown types.
+           ;; Unfortunately, people love to suppress COMPILER-NOTE because SBCL produces
+           ;; far too many of those for low-level things like untagged-SAP-to-tagged-SAP.
+           ;; So we could opt to STYLE-WARN, which is, in this case, perhaps more severe
+           ;; than we'd like?
+           ;; I guess we're just going to have to say that if you've muffled too may
+           ;; kinds of NOTEs, that's on you.
+           #-sb-xc-host (compiler-notify 'unknown-typep-note
+                                         :format-control "can't open-code test of unknown type ~S"
+                                         :format-arguments (list (type-specifier type)))
            `(let ((object ,object)
                   (cache (load-time-value (cons #'sb-kernel::cached-typep ',spec)
                                           t)))
@@ -599,7 +613,7 @@
                 `(characterp ,object))))
         (let ((n-code (sb-xc:gensym "CODE")))
           `(and (characterp ,object)
-                (let ((,n-code (sb-xc:char-code ,object)))
+                (let ((,n-code (char-code ,object)))
                   (or
                    ,@(loop for pair in pairs
                            collect
@@ -825,107 +839,71 @@
 ;;; object's layout can ever be EQ to that of the ancestor.
 ;;; e.g. a fixnum as representative of class REAL.
 ;;; So in actual practice, you can't make something that is a pure STREAM, etc.
-(defvar *use-layout-ids* (or #+(or ppc64 x86 x86-64) t))
+#-(or x86 x86-64) ; vop-translated for these 2
+(defmacro layout-depthoid-ge (layout depthoid)
+  `(>= (wrapper-depthoid ,layout) ,depthoid))
+(symbol-macrolet ((get-hash #+metaspace 'layout-clos-hash #-metaspace 'wrapper-clos-hash)
+                  (get-flags #+metaspace 'layout-flags #-metaspace 'wrapper-flags))
 (defun transform-instance-typep (classoid)
   (binding*
       ((name (classoid-name classoid))
-       (layout (let ((res (info :type :compiler-layout name)))
-                 (when (and res (not (layout-invalid res))) res)))
-       ((primtype-predicate slot-reader)
+       (wrapper (let ((res (info :type :compiler-layout name)))
+                 (when (and res (not (wrapper-invalid res))) res)))
+       (layout (and wrapper (wrapper-friend wrapper)))
+       ((lowtag lowtag-test slot-reader)
         (cond ((csubtypep classoid (specifier-type 'funcallable-instance))
-               (values '(function-with-layout-p object)
-                       '(%fun-layout object)))
-              ((or (csubtypep classoid (specifier-type 'instance))
-                   (eq classoid (specifier-type 'logical-pathname))
-                   ;; CONDITION can't be a funcallable-instance
-                   (csubtypep classoid (specifier-type 'condition)))
-               (values '(%instancep object)
-                       '(%instance-layout object)))))
-       (get-layout-or-return-false
-        (if primtype-predicate
-            ;; Test just one of %INSTANCEP or %FUNCALLABLE-INSTANCE-P
-            `(if ,primtype-predicate ,slot-reader (return-from typep nil))
-            ;; But if we don't know which is will be, try both.
-            ;; This is less general than LAYOUT-OF,and therefore
-            ;; a little quicker to fail, because objects with
-            ;; {LIST|OTHER}-POINTER-LOWTAG can't possibly pass.
-            ;; It's a bit disappointing that STREAM uses this slower path,
-            ;; but some people think it should be possible to create
-            ;; funcallable streams. As a countermeasure, it would be possible to emit
-            ;; slightly better code in a vop.
-            `(cond ((%instancep object) (%instance-layout object))
-                   ((function-with-layout-p object) (%fun-layout object))
-                   (t (return-from typep nil)))))
-       (depthoid (if layout (layout-depthoid layout) -1))
-       (n-layout (make-symbol "LAYOUT"))
-       (n-inherits (make-symbol "INHERITS"))
-        ;; In order to efficiently perform the DEEPER-P test without this hack of using
-        ;; a vop (when available for non-risc machines), we'd have to do two things:
-        ;; - have instcombine combine the read and compare as one instruction
-        ;; - implement half-width structure slots
-        ;; Since both of those are not happening any time soon, ...
-       (deeper-p
-        (if (vop-existsp :translate layout-depthoid-gt)
-            `(layout-depthoid-gt ,n-layout ,depthoid)
-            `(> (layout-depthoid ,n-layout) ,depthoid)))
-       (nth-ancestor ; This is possibly unused (if no compile-time layout, or depthoid -1)
-        ;; Use DATA-VECTOR-REF directly, since that's what SVREF in SAFETY 0 will become.
-        `(locally (declare (optimize (safety 0)))
-           (data-vector-ref (layout-inherits ,n-layout) ,depthoid)))
-       (ancestor-layout-eq
-        `(eq ,nth-ancestor ,layout)))
+               (values sb-vm:fun-pointer-lowtag
+                       '(function-with-layout-p object) '(%fun-layout object)))
+              ((csubtypep classoid (specifier-type 'instance))
+               (values sb-vm:instance-pointer-lowtag
+                       '(%instancep object) '(%instance-layout object)))))
+       (depthoid (if wrapper (wrapper-depthoid wrapper) -1))
+       (type (make-symbol "TYPE")))
+    (declare (ignorable layout))
 
     ;; Easiest case first: single bit test.
     (cond ((member name '(condition pathname structure-object))
            `(and (%instancep object)
-                 (logtest (layout-flags (%instance-layout object))
+                 (logtest (,get-flags (%instance-layout object))
                           ,(case name
                              (condition +condition-layout-flag+)
                              (pathname  +pathname-layout-flag+)
                              (t         +structure-layout-flag+)))))
 
-          ;; Next easiest: Sealed and at most one subclass.
-          ;; However, it is better not to perform two IFs when a single one
-          ;; will suffice, so forbid any subclass when using layout-ids.
-          ((and layout
+          ;; TODO: remove after April 2021 release.
+          ((eq name 'sb-kernel::random-class)
+           (style-warn "~S should not appear in a TYPEP test" name)
+           nil)
+
+          ;; Next easiest: Sealed and no subtypes. Typically for DEFSTRUCT only.
+          ;; Even if you don't seal a DEFCLASS, we're allowed to assume that things
+          ;; won't change, as per CLHS 3.2.2.3 on Semantic Constraints:
+          ;;  "Classes defined by defclass in the compilation environment must be defined
+          ;;  at run time to have the same superclasses and same metaclass."
+          ;; I think that means we should know the lowtag always. Nonetheless, this isn't
+          ;; an important scenario, and only if you _do_ seal a class could this case be
+          ;; reached; users rarely seal their classes since the standard doesn't say how.
+          ((and wrapper
                 (eq (classoid-state classoid) :sealed)
-                (or (not (classoid-subclasses classoid))
-                    (and (not *use-layout-ids*)
-                         (eql (hash-table-count (classoid-subclasses classoid))
-                              1))))
-            ;; It's possible to seal a STANDARD-CLASS, not just a STRUCTURE-CLASS,
-            ;; though probably extremely weird. Also the PRED should be set in
-            ;; that event, but it isn't.
-            ;; The crummy dual expressions for the same result are because
-            ;; (BLOCK (RETURN ...)) seems to emit a forward branch in the
-            ;; passing case, but AND emits a forward branch in the failing
-            ;; case which I believe is the better choice.
-            (let ((other-layout (and (classoid-subclasses classoid)
-                                     (dohash ((classoid layout)
-                                              (classoid-subclasses classoid)
-                                              :locked t)
-                                             (declare (ignore classoid))
-                                             (return layout)))))
-              (flet ((check-layout (layout-getter)
-                       (cond (other-layout
-                              ;; It's faster to compare two layouts than
-                              ;; doing whatever is done below
-                              `(let ((object-layout ,layout-getter))
-                                 (or (eq object-layout ',layout)
-                                     (eq object-layout ',other-layout))))
-                             ((and (vop-existsp :named sb-vm::layout-eq)
-                                   (equal layout-getter '(%instance-layout object)))
-                              `(sb-vm::layout-eq object ',layout))
-                             (t
-                              `(eq ,layout-getter ',layout)))))
-                (if primtype-predicate
-                    `(and ,primtype-predicate ,(check-layout slot-reader))
-                    `(block typep ,(check-layout get-layout-or-return-false))))))
+                (not (classoid-subclasses classoid)))
+           (if lowtag-test
+               `(and ,lowtag-test
+                     ,(if (vop-existsp :translate layout-eq)
+                          `(layout-eq object ,wrapper ,lowtag)
+                          `(eq ,slot-reader ,layout)))
+               ;; `(eq ,layout
+               ;;      (if-vop-existsp (:translate %instanceoid-layout)
+               ;;        (%instanceoid-layout object)
+               ;;        ;; Slightly quicker than LAYOUT-OF. See also %PCL-INSTANCE-P
+               ;;        (cond ((%instancep object) (%instance-layout object))
+               ;;              ((funcallable-instance-p object) (%fun-layout object))
+               ;;              (t ,(find-layout 't)))))
+               (bug "Unexpected metatype for ~S" wrapper)))
 
           ;; All other structure types
-          ((and (typep classoid 'structure-classoid) layout)
+          ((and (typep classoid 'structure-classoid) wrapper)
             ;; structure type tests; hierarchical layout depths
-            (aver (equal primtype-predicate '(%instancep object)))
+            (aver (eql lowtag sb-vm:instance-pointer-lowtag))
             ;; we used to check for invalid layouts here, but in fact that's both unnecessary and
             ;; wrong; it's unnecessary because structure classes can't be redefined, and it's wrong
             ;; because it is quite legitimate to pass an object with an invalid layout
@@ -934,57 +912,46 @@
                     ;; If we allowed structure classes to be mixed in to standard-object,
                     ;; this might have to change to consider object invalidation. Probably would
                     ;; want to track structure classoids that would render this code inadmissible.
-                  (let ((,n-layout (%instance-layout object)))
-                    ,(if *use-layout-ids*
-                         (if (<= depthoid sb-kernel::layout-id-vector-fixed-capacity)
-                             `(%structure-is-a ,n-layout ,layout)
-                             `(and (layout-depthoid-ge ,n-layout ,depthoid)
-                                   (%structure-is-a ,n-layout ,layout)))
-                         ;; Distinguish between abstract base types - no DD-CONSTRUCTOR,
-                         ;; hence no direct instances - and everything else.
-                         (cond ((dd-constructors (layout-dd layout))
-                                `(cond ((eq ,n-layout ,layout) t)
-                                       (,deeper-p ,ancestor-layout-eq)))
-                               (t ; abstract base type deeper than optimized max.
-                                ;; Assume that no layout is EQ to the base layout,
-                                ;; and unconditionally fetch and dereference layout-inherits.
-                                `(eq (if ,deeper-p ,nth-ancestor ,n-layout) ,layout)))))))
+                  ,(if (<= depthoid sb-kernel::layout-id-vector-fixed-capacity)
+                       `(%structure-is-a (%instance-layout object) ,wrapper)
+                       `(let ((,type (%instance-layout object)))
+                          (and (layout-depthoid-ge ,type ,depthoid)
+                               (%structure-is-a ,type ,wrapper))))))
 
-          ((> depthoid 0) ; fixed-depth ancestors of non-structure types: STREAM, FILE-STREAM,
-           ;; SEQUENCE; all are abstract base types.
-           ;; (I don't know why sometimes we just call SEQUENCEP and STREAMP. Policy?)
+          ((> depthoid 0)
+           ;; fixed-depth ancestors of non-structure types:
+           ;; STREAM, FILE-STREAM, STRING-STREAM, and SEQUENCE.
             #+sb-xc-host (when (typep classoid 'static-classoid)
                            ;; should have use :SEALED code above
-                           (bug "Non-frozen static classoids?"))
-            ;; There is no need to test the inheritance depth for depthoid 1 because INHERITS
-            ;; will contains at least #<LAYOUT for T> and a padding word if naught else.
-            ;; Though in actual fact it can't have just T, because users have no means to create
-            ;; types that are direct descendants of type T.
-            ;; However, we do have to avoid unsafely dereferencing inherits for FILE-STREAM
-            ;; which is at depth 4. An easy way to avoid this depth test would be requiring at
-            ;; least 5 elements in any INHERITS vector.  Alternatively, adding 1 bit each for
-            ;; SEQUENCE, STREAM, and FILE-STREAM to LAYOUT-FLAGS would suffice.
-            (let ((guts `(,@(unless (eq name 'condition)
-                              `((when (zerop (layout-clos-hash ,n-layout))
-                                 (setq ,n-layout
-                                  (truly-the layout
-                                             (update-object-layout-or-invalid
-                                              object ',layout))))))
-                          ,(if (eql depthoid 1)
-                               ancestor-layout-eq
-                               `(let ((,n-inherits (layout-inherits ,n-layout)))
-                                  (and (> (length ,n-inherits) ,depthoid)
-                                       (eq (data-vector-ref ,n-inherits ,depthoid)
-                                           ,layout)))))))
-
-              (if primtype-predicate
-                  `(and ,primtype-predicate (let ((,n-layout ,slot-reader)) ,@guts))
-                  `(block typep
-                     (let ((,n-layout ,get-layout-or-return-false)) ,@guts)))))
+                           (bug "Non-frozen static classoids ~S" name))
+            (let ((guts `((when (zerop (,get-hash ,type))
+                            (setq ,type (update-object-layout object)))
+                          ,(ecase name
+                            (stream
+                             `(logtest (,get-flags ,type) ,+stream-layout-flag+))
+                            (file-stream
+                             `(logtest (,get-flags ,type) ,+file-stream-layout-flag+))
+                            (string-stream
+                             `(logtest (,get-flags ,type) ,+string-stream-layout-flag+))
+                            ;; Testing the type EXTENDED-SEQUENCE tests for #<LAYOUT of SEQUENCE>.
+                            ;; It can only arise from a direct invocation of TRANSFORM-INSTANCE-TYPEP,
+                            ;; because the lisp type is not a classoid. It's done this way to define
+                            ;; the logic once only, instead of both here and src/code/pred.lisp.
+                            (sequence
+                             `(logtest (,get-flags ,type) ,+sequence-layout-flag+))))))
+              (if lowtag-test
+                  `(and ,lowtag-test (let ((,type ,slot-reader)) ,@guts))
+                  (if-vop-existsp (:translate %instanceoid-layout)
+                    `(let ((,type (%instanceoid-layout object))) ,@guts)
+                    `(block typep
+                       (let ((,type (cond ((%instancep object) (%instance-layout object))
+                                          ((funcallable-instance-p object) (%fun-layout object))
+                                          (t (return-from typep nil)))))
+                         ,@guts))))))
 
           (t
             `(classoid-cell-typep ',(find-classoid-cell name :create t)
-                                  object)))))
+                                  object))))))
 
 ;;; If the specifier argument is a quoted constant, then we consider
 ;;; converting into a simple predicate or other stuff. If the type is
@@ -1298,49 +1265,57 @@
                     `(the ,result-type
                           (complex (coerce (realpart x) ',part-type)
                                    (coerce (imagpart x) ',part-type)))))))))
-      ;; Special case STRING and SIMPLE-STRING as they are union types
-      ;; in SBCL.
-      ((member tval '(string simple-string))
-       `(the ,tval
-             (if (typep x ',tval)
-                 x
-                 (replace (make-array (length x) :element-type 'character) x))))
       ((eq tval 'character)
        `(character x))
-      ;; Special case VECTOR
-      ((eq tval 'vector)
-       `(the ,tval
-             (if (vectorp x)
-                 x
-                 (replace (make-array (length x)) x))))
       ;; Handle specialized element types for 1D arrays.
-      ((csubtypep tspec (specifier-type '(array * (*))))
-       ;; Can we avoid checking for dimension issues like (COERCE FOO
-       ;; '(SIMPLE-VECTOR 5)) returning a vector of length 6?
-       ;;
-       ;; CLHS actually allows this for all code with SAFETY < 3,
-       ;; but we're a conservative bunch.
-       (if (or (policy node (zerop safety)) ; no need in unsafe code
-               (and (array-type-p tspec) ; no need when no dimensions
-                    (equal (array-type-dimensions tspec) '(*))))
-           ;; We can!
-           (multiple-value-bind (vtype etype upgraded) (simplify-vector-type tspec)
-             (unless upgraded
-               (give-up-ir1-transform))
-             (let ((vtype (type-specifier vtype)))
-               `(the ,vtype
-                     (if (typep x ',vtype)
-                         x
-                         (replace
-                          (make-array (length x)
-                                      ,@(and (not (eq etype *universal-type*))
-                                             (not (eq etype *wild-type*))
-                                             `(:element-type ',(type-specifier etype))))
-                          x)))))
-           ;; No, duh. Dimension checking required.
-           (give-up-ir1-transform
-            "~@<~S specifies dimensions other than (*) in safe code.~:@>"
-            tval)))
+      ((multiple-value-bind (result already-type-p dimension specialization)
+           (cond ((or (and (array-type-p tspec)
+                           (neq (array-type-complexp tspec) t) ; :MAYBE and NIL are good
+                           (not (contains-unknown-type-p (array-type-element-type tspec)))
+                           ;; just for requesting (array nil (*)), you lose
+                           (neq (array-type-specialized-element-type tspec) *empty-type*)
+                           (consp (array-type-dimensions tspec))))
+                  (values tspec
+                          (source-transform-array-typep 'x tspec)
+                          (car (array-type-dimensions tspec))
+                          (let ((et (array-type-specialized-element-type tspec)))
+                            (unless (or (eq et *universal-type*) ; don't need
+                                        ;; * is illegal as :element-type; in this context
+                                        ;; it means to produce a SIMPLE-VECTOR
+                                        (eq et *wild-type*))
+                              `(:element-type ',(type-specifier et))))))
+                 ;; Check for string types.  This loses on (STRING 1) and such.
+                 #+sb-unicode
+                 ((type= tspec (specifier-type 'simple-string))
+                  (values 'simple-string '(simple-string-p x) '* '(:element-type 'character)))
+                 #+sb-unicode
+                 ((type= tspec (specifier-type 'string))
+                  (values 'string '(stringp x) '* '(:element-type 'character))))
+         (when result
+           ;; If the dimension is in the type, we check the input length if safety > 0,
+           ;; though technically CLHS would allow not checking in safety < 3.
+           ;; And if mismatch occurs in unsafe code, the results accords with the
+           ;; specifier, NOT the dimension of the input. This is a rational choice
+           ;; because one could not argue that incorrect code should have taken the
+           ;; bad input's length when COERCE was asked for an exact type of output.
+           `(truly-the ,result
+               (if ,already-type-p
+                   x
+                   ,(cond ((eq dimension '*)
+                           #+ubsan
+                           ;; Passing :INITIAL-CONTENTS avoids allocating ubsan shadow bits,
+                           ;; but redundantly checks the length of the input in MAKE-ARRAY's
+                           ;; transform because we don't or can't infer that LENGTH gives the
+                           ;; same answer each time it is called on X. There may be a way to
+                           ;; extract more efficiency - at least eliminate the unreachable
+                           ;; error-signaling code on mismatch - but I don't care to try.
+                           `(make-array (length x) ,@specialization :initial-contents x)
+                           #-ubsan ; better: do not generate a redundant LENGTH check
+                           `(replace (make-array (length x) ,@specialization) x))
+                          ((policy node (= safety 0)) ; Disregard the input length
+                           `(replace (make-array ,dimension ,@specialization) x))
+                          (t
+                           `(make-array ,dimension ,@specialization :initial-contents x))))))))
       ((type= tspec (specifier-type 'list))
        `(coerce-to-list x))
       ((csubtypep tspec (specifier-type 'function))
@@ -1372,3 +1347,6 @@
           ((csubtypep (lvar-type object) this-type)
            nil)
           ((give-up-ir1-transform)))))
+
+;;; BIGNUMP is simpler than INTEGERP, so if we can rule out FIXNUM then ...
+(deftransform integerp ((x) ((not fixnum)) * :important nil) '(bignump x))
